/*
 * libtrackle.c
 *
 * This file defines libtrackle's tracking functions and other
 * helper functions used by libtrackle. No overridden library
 * functions should appear in this file.
 *
 * 2006 Oct 5
 * Dan Crosta <dcrosta@sccs.swarthmore.edu>
 *
 * based on Audlib by Benjamin Kuperman and Mustafa Paksoy
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <dlfcn.h>
#include <sys/times.h>
#include <errno.h>

#include "protos.h"
#include "paths.h"


/*------------------------------------------------------------------------
 * macros
 *------------------------------------------------------------------------
 */
void __attribute__ ((constructor)) libtrackle_init(void);
void __attribute__ ((destructor))  libtrackle_fini(void);


/*------------------------------------------------------------------------
 * Global Variables
 *------------------------------------------------------------------------
 */
int libtrackle_initialized=0;
int trackle_UID;
char *trackle_sessionDir, *trackle_fileCounter;


/*------------------------------------------------------------------------
 * init and fini
 *------------------------------------------------------------------------
 */
void libtrackle_init() {
    char *err;
    char *tempUID;
   
    /* set up local function pointers for libc functions */
    trackle_funcptr_setup();


    // cache TRACKLE_SESSION_DIR, TRACKLE_FILE_COUNTER, and TRACKLE_UID
    trackle_sessionDir = getenv("TRACKLE_SESSION_DIR"); 
    if (NULL == trackle_sessionDir) {
        msg("TRACKLE_SESSION_DIR environment variable not set...quitting.\n");
        exit(-1);
    }
    trackle_sessionDir = strdup(trackle_sessionDir);

    trackle_fileCounter = getenv("TRACKLE_FILE_COUNTER"); 
    if (NULL == trackle_fileCounter) {
        msg("TRACKLE_FILE_COUNTER environment variable not set...quitting.\n");
        exit(-1);
    }
    trackle_fileCounter = strdup(trackle_fileCounter);
    
    tempUID = getenv("TRACKLE_UID"); 
    if (NULL == tempUID) {
        msg("TRACKLE_UID environment variable not set...quitting.\n");
        exit(-1);
    }
    trackle_UID = strtol(tempUID, &err, 10);
    if ( (*__strlenptr)(tempUID) != (err - tempUID) ) {
        msg("improperly formatted TRACKLE_UID environment variable...quitting.\n");
        exit(-1);
    }

    // don't need to free tempUID since getenv()
    // does not copy the string, just returns
    // a pointer to the string in the environment

    libtrackle_initialized=1;
}

void libtrackle_fini() {
    // Get end versions of files
    trackle_logFini();
    
    // clean up after environment variable-snatching
    (*__freeptr)(trackle_sessionDir);
    (*__freeptr)(trackle_fileCounter);

    libtrackle_initialized=0;
}



/* trackle_logFileAccess
 *
 * The main driver.  This method should always do the Right Thing with regards
 * to logging file accesses.  Right now, it gives each file a unique ID for the
 * session based on a counter of the number of files indexed (eg the first file
 * indexed is ID 00000, the second is ID 00001, etc.), builds the correct paths
 * to the files to be written, and then calls methods to trackle_makeIndexEntry,
 * trackle_copyFile, and trackle_statFile.  The index is stored in the base session directory
 * and the file and stat are stored in a files/ subdirectory as ID.file and
 * ID.stat respectively.
 */
int trackle_logFileAccess(const char path[]){
    char indexPathL[PATH_MAX], filePathL[PATH_MAX], statPathL[PATH_MAX];
    int ID;
    char test[PATH_MAX];
    int firstAccess;
    int exists = 0;
    int readable = 0;
    struct stat statbuf;

#ifdef DEBUG
    char _path[PATH_MAX];
    int len = 0;

    len = (*__readlinkptr)("/proc/self/exe", _path, PATH_MAX);
    _path[len] = 0;

    
    (*__fprintfptr)(stderr, "logFileAccess(%s) by %s\n", path, _path);

#endif

    if ( !libtrackle_initialized ) {
        msg("trackle_logFileAccess called with before libtrackle initialization!\n");
        return -1;
    }
    
    // Do some simple path expansion
    if( (path[0] == '.') && (path[1] == '/') ){
        getcwd(test, PATH_MAX);
        strcat(test, "/");
        strcat(test, path+2);
        path = test;
    } else if( path[0] != '/' ){
        getcwd(test, PATH_MAX);
        strcat(test, "/");
        strcat(test, path);
        path = test;
    }


    if ( -1 == trackle_stat(path, &statbuf) ) {
        // stat
        switch (errno) {
        case EACCES:
        case ENOTDIR:
            // dealing with a funky path or a path we
            // can't see the file on, so just return
            return 0;

        case ENOENT:
            // file doesn't exist, probably
            break;
        }
    }
    

    
    // check if the file exists
    exists = (0 == access(path, F_OK));

    // check if the file is readable
    readable = (0 == access(path, R_OK));

    
    // if it exists and is *not* a regular file,
    // quit now without doing anything
    if ( exists && !(S_ISREG(statbuf.st_mode)) ) {
        return 0;
    }

    

    // set up index paths
    (*__snprintfptr)(indexPathL, PATH_MAX, "%s/index", trackle_sessionDir);

    ID = trackle_alreadyLogged(path, indexPathL);
    if( -2 == ID ){
        firstAccess = 1;

        
        // get the file index number 
        // (We need to make sure that each file is uniquely identified.  We don't
        // want /foo/bar.conf and /bar.conf to collide.)
        ID = trackle_getNumFiles();

        // set up log paths
        (*__snprintfptr)(filePathL, PATH_MAX, "%s/files/%05d.file", trackle_sessionDir, ID);
        (*__snprintfptr)(statPathL, PATH_MAX, "%s/files/%05d.stat", trackle_sessionDir, ID);

    } else {
        firstAccess = 0;
        (*__snprintfptr)(filePathL, PATH_MAX, "%s/files/%05d.file.end", trackle_sessionDir, ID);
        (*__snprintfptr)(statPathL, PATH_MAX, "%s/files/%05d.stat.end", trackle_sessionDir, ID);
    }

    
    
    
    // if it's the first access, make the index entry
    if (1 == firstAccess){
        trackle_makeIndexEntry(ID, path, indexPathL);
    }


    // always stat the file
    trackle_statFile(path, statPathL, exists, readable);

    
    if ( exists && readable ) {
        trackle_copyFile(path, filePathL);
    }
    
    return 0;
}


/* trackle_makeIndexEntry
 * 
 * make a comma separated entry in the index file consisting of file ID and path
 * 
 */
int trackle_makeIndexEntry(int index, const char path[], char indexPath[]){
    FILE *indexFile;
   
    // open index file
    if((indexFile = (*__fopenptr)(indexPath, "a")) == NULL){
        (*__fprintfptr)(stderr, "failed to open index file `%s` in trackle_makeIndexEntry\n", indexPath);
        perror("fopen");
        return -1;
    }

    (*__fprintfptr)(indexFile, "%s,%05d\n", path, index);

    // close index file carefully
    if( (*__fcloseptr)(indexFile) == EOF ){
        printf("failed to close index file `%s` in trackle_makeIndexEntry\n", indexPath);
    }

    return 0;
}

/* trackle_copyFile
 *
 * Copies the contents of one file to another.  If the destination file
 * exists, it is overwritten.
 */
int trackle_copyFile(const char sourcePath[], char destPath[]){
    FILE *sourceFile, *destFile;
    char ch;
  
    // open source and destination files
    if((sourceFile = (*__fopenptr)(sourcePath, "r")) == NULL){    
        printf("failed to open source file `%s` in trackle_copyFile\n", sourcePath);
        perror("fopen");
        // some day this should fail gracefully. but really, it just shouldn't.
        return -1;
    }
   if((destFile = (*__fopenptr)(destPath, "w")) == NULL){
        printf("failed to open dest file `%s` in trackle_copyFile\n", destPath);
        perror("fopen");
        // some day this should fail gracefully. but really, it just shouldn't.
        return -1;
    }

   // copy source to destination
    while( (ch = fgetc(sourceFile)) != EOF ){
        fputc(ch, destFile);
    }

    // close source and destination files carefully
    if( (*__fcloseptr)(sourceFile) == EOF ){
        printf("failed to close source file `%s` in trackle_copyFile\n", sourcePath);
        perror("fclose");
    }
    if( (*__fcloseptr)(destFile) == EOF ){
        printf("failed to close dest file `%s` in trackle_copyFile\n", destPath);
        perror("fclose");
    }
    // Make sure only the owner can read and write the file (for security)
    (*__chownptr)(destPath, trackle_UID, 0);
    (*__chmodptr)(destPath, S_IRUSR | S_IWUSR);

    return 0;
}

/* trackle_statFile
 * 
 * Stats the file in question and records all stat information in a .stat file.
 *
 * If the file exists, the .stat (or .stat.end) file will contain each of the fields
 * from the struct stat structure, plus a 'special:' field. This field contains
 * space-separated tokens for some special conditions: 'noexist' if the file in
 * question does not exist, and 'unreadable(uid)' if the file is not readable by
 * the current user (who has uid "uid").
 */
int trackle_statFile(const char sourcePath[], char destPath[], int exists, int readable){
    FILE *destFile;
    struct stat statBuf;

    // open destination file for writing
    if((destFile = (*__fopenptr)(destPath, "w")) == NULL){
        printf("failed to open dest file `%s` in trackle_statFile\n", destPath);
        perror("fopen");
        return -1;
    }
   
    if ( exists ) {
        // stat file
        if(trackle_stat(sourcePath, &statBuf) == -1){
            printf("failed to stat file `%s` in trackle_statFile\n", sourcePath);
            perror("stat");
            return -1;
        }

        // write stat info to dest file
        (*__fprintfptr)(destFile, "dev: %d\n",     (int)statBuf.st_dev);
        (*__fprintfptr)(destFile, "inode: %d\n",   (int)statBuf.st_ino);
        (*__fprintfptr)(destFile, "mode: %o\n",    statBuf.st_mode);
        (*__fprintfptr)(destFile, "nlink: %d\n",   statBuf.st_nlink);
        (*__fprintfptr)(destFile, "uid: %d\n",     statBuf.st_uid);
        (*__fprintfptr)(destFile, "gid: %d\n",     statBuf.st_gid);
        (*__fprintfptr)(destFile, "rdev: %d\n",    (int)statBuf.st_rdev);
        (*__fprintfptr)(destFile, "size: %d\n",    (int)statBuf.st_size);
        (*__fprintfptr)(destFile, "blksize: %d\n", (int)statBuf.st_blksize);
        (*__fprintfptr)(destFile, "blocks: %d\n",  (int)statBuf.st_blocks);
        (*__fprintfptr)(destFile, "mtime: %d\n",   (int)statBuf.st_mtime);
        (*__fprintfptr)(destFile, "ctime: %d\n",   (int)statBuf.st_ctime);
    }

    // write the special fields
    (*__fprintfptr)(destFile, "special: ");

    if ( !exists )
        (*__fprintfptr)(destFile, "noexist ");
    if ( !readable )
        (*__fprintfptr)(destFile, "unreadable(%d) ", geteuid());

    (*__fprintfptr)(destFile, "\n");


    // close destination file carefully
    if( (*__fcloseptr)(destFile) == EOF ){
        printf("failed to close dest file `%s` in trackle_statFile\n", destPath);
    }

    // Make sure only the owner can read and write the file (for security)
    // (can only do this as root)
    if ( geteuid() == 0 ) {
        (*__chownptr)(destPath, trackle_UID, 0);
        (*__chmodptr)(destPath, S_IRUSR | S_IWUSR);
    }

    return 0;
}

/* trackle_logFini
 * 
 * For each file in the index, we log another file access to make sure we have
 * the latest version.  The new .end files will overwrite the old ones. 
 */
int trackle_logFini(){
    char indexPathL[PATH_MAX];
    FILE *indexFile;
    char path[PATH_MAX], ID[6];
    
    // set up index path
    (*__snprintfptr)(indexPathL, PATH_MAX, "%s/index", trackle_sessionDir);
    
    // open index file for reading
    if((indexFile = (*__fopenptr)(indexPathL, "r")) == NULL){    
        (*__fprintfptr)(stderr, "failed to open index file `%s` in trackle_logFini\n", indexPathL);
        perror("fopen");
        return -1;
    }
    
    // for each file in the index, trackle_logFileAccess
    while( 0 == trackle_getNextIndexLine(indexFile, path, ID)){
        trackle_logFileAccess(path);
    }
    
    // close index file carefully
    if( (*__fcloseptr)(indexFile) == EOF ){
        (*__fprintfptr)(stderr, "failed to close index file `%s` in trackle_logFini\n", indexPathL);
    }
    
    return 0;
}

/* trackle_getNextIndexLine
 *
 * Iterates over lines in an index file of the form "<path>,<ID>\n".
 *
 * returns: 0 if it has filled path and ID, 1 if there are no more lines
 */
int trackle_getNextIndexLine(FILE *indexFile, char *path, char *ID){
    int flag = 0;
    int pathCount = 0;
    int IDCount = 0;
    char ch;

    // 
    while( (ch = fgetc(indexFile)) != EOF ){
        if(flag == 0){
            if(ch == ','){
                flag = 1;
                continue;
            }
            path[pathCount] = ch;
            path[pathCount+1] = '\0';
            pathCount++;
        } else if(flag == 1){
            if(ch == '\n'){
                return 0;
            }
            ID[IDCount] = ch;
            ID[IDCount+1] = '\0';
            IDCount++;
        }
    }

    return 1;
}

/* trackle_alreadyLogged
 *
 * Determines if the file <path> has already been logged in the file
 * <index>.
 *
 * Returns ID if already logged, -2 if not, and -1 on other error.
 */
int trackle_alreadyLogged(const char path[], char index[]){
    FILE *indexFile;
    char ch;
    char line[PATH_MAX+7], ID[6];
    int i;
    char* err;
    int val;

    // open index file
    if((indexFile = (*__fopenptr)(index, "r")) == NULL){   
        (*__fprintfptr)(stderr, "failed to open index file `%s` in trackle_alreadyLogged\n", index);
        perror("fopen");
        return -1;
    }

    // check line-by-line to see if the file is already open
    i = 0;
    while ( (ch = fgetc(indexFile)) != EOF ){
        if ( ch == '\n' ) {
            line[i] = '\0';

            // compare the first N characters of this line
            // to the value of the new path [for N == strlen(path)]
            if( (strncmp(path, line, (*__strlenptr)(path)) == 0) && (line[(*__strlenptr)(path)] == ',') ){

                // Close index file
                if( (*__fcloseptr)(indexFile) == EOF ){
                    (*__fprintfptr)(stderr, "failed to close index file\n");
                }

                // XXX: what if more than 10k files are changed in a session?
                (*__strncpyptr)(ID, &(line[(*__strlenptr)(path) + 1]), 5);
                ID[5] = '\0';

                val = strtol( ID, &err, 10 );
                
                if ( err != ID + strlen(ID) ) {
                    (*__fprintfptr)(stderr, "Could not parse index value `%s' as integer\n", ID);
                    return -1;
                }
                return val;
            }
            i = 0;
        } else {
            line[i] = ch;
            i++;
        }   
    }
 
    // Close index file
    if( (*__fcloseptr)(indexFile) == EOF ){
        (*__fprintfptr)(stderr, "failed to close index file\n");
    }

    return -2;
}
    
/* trackle_getNumFiles
 *
 * We need to keep track of the number of files that have been opened, mostly as
 * a means for assigning unique index numbers.  Because libtrackle is loaded in
 * a separate instance every time and we need to share the number of files
 * among all instances within a given session, the state will be shared in a
 * counter file.  Each file is indicated by a newline character.  This can be
 * easily counted.
 *
 * TODO: We need mutual exclusion of this shared resource.
 */
int trackle_getNumFiles(){
    FILE *cfile;
    int counter = 0;
    //int semid;
 
    if((cfile = (*__fopenptr)(trackle_fileCounter, "r")) == NULL){    
        printf("failed to open counter file `%s` in trackle_getNumFiles\n", trackle_fileCounter);
        perror("fopen");
        return -1;
    }

    while(fgetc(cfile) != EOF){
        counter++;
    }
    
    if( (*__fcloseptr)(cfile) == EOF ){
        printf("failed to close counter file `%s` in trackle_getNumFiles\n", trackle_fileCounter);
    }
   
    trackle_incNumFiles();
    
    //semop(semid, &sem_up, 1);
    
    return counter;
}

/* trackle_incNumFiles
 *
 * Appends a newline to the counter file.
 */
int trackle_incNumFiles(){
    FILE *cfile;
    
    if((cfile = (*__fopenptr)(trackle_fileCounter, "a")) == NULL){    
        printf("failed to open counter file `%s` in trackle_incNumFiles\n", trackle_fileCounter);
        perror("fopen");
        // some day this should fail gracefully. but really, it just shouldn't.
        return -1;
    }

    fputc('\n', cfile);
    
    if( (*__fcloseptr)(cfile) == EOF ){
        printf("failed to close counter file `%s` in trackle_incNumFiles\n", trackle_fileCounter);
    }

    return 0;
}

/* makes sure that LD_PRELOAD, TRACKLE_SESSION_DIR, TRACKLE_FILE_COUNTER, and
 * TRACKLE_UID are set in the environment. should only be used from within the
 * interposed execv() and execvp() functions
 */
void trackle_fixenv(void) {
    char uidStr[1024];

    // reset LD_PRELOAD to be just 'libtrackle.so'
    unsetenv("LD_PRELOAD");
    putenv("LD_PRELOAD=libtrackle.so");  //XXX: if we ever make the name of the
                                         //library configurable, we need to change this!

    // preapre a string version of trackle_UID for putting into the environment
    (*__snprintfptr)(uidStr, 1023, "%d", trackle_UID);
    uidStr[1023] = '\0';

    // reset TRACKLE_SESSION_DIR, TRACKLE_FILE_COUNTER
    setenv("TRACKLE_SESSION_DIR",  trackle_sessionDir, 1);
    setenv("TRACKLE_FILE_COUNTER", trackle_fileCounter, 1);
    setenv("TRACKLE_UID",          uidStr, 1);
}



/*------------------------------------------------------------------------
 * trackle_libopen - open a dynamic library
 *------------------------------------------------------------------------
 */
void* trackle_libopen(const char *path) {
    void *ptr;
    if (NULL == (ptr=dlopen(path,RTLD_LAZY))) {
        fprintf(stderr, "Problem with dlopen: %s\n", dlerror());
    }
    return(ptr);
}


/*------------------------------------------------------------------------
 * trackle_getfunc - search for the next library function
 *------------------------------------------------------------------------
 */
void* trackle_getfunc(void *handle, const char *name) {
    if (!handle) {
        fprintf(stderr, "Passed a null pointer looking up: %s\n", name);

        /* TODO - figure out how to handle this case correctly */
        return NULL;
    }

    return dlsym(handle,name);
}


ssize_t (*__writeptr)(int , const void *, size_t )=NULL;
ssize_t (*__readptr)(int , void *, size_t )=NULL;
size_t  (*__fwriteptr)(const void *, size_t, size_t, FILE *)=NULL;
int     (*__openptr)(const char *, int , /* mode_t mode */...)=NULL;
int     (*__closeptr)(int )=NULL;
int     (*__fcloseptr)(FILE *)=NULL;
int     (*__vprintfptr)(const char *, va_list)=NULL;
int     (*__printfptr)(const char *, /* args*/ ...)=NULL; 
int     (*__fprintfptr)(FILE *, const char *, /* args*/ ...)=NULL;
int     (*__snprintfptr)(char *, size_t, const char *, /* args*/ ...)=NULL;
int     (*__vsnprintfptr)(char *, size_t, const char *, va_list)=NULL;
#ifdef LINUX
int     (*__statptr)(int,const char *, struct stat *)=NULL;
int     (*__lstatptr)(int,const char *, struct stat *)=NULL;
#else
int     (*__lstatptr)(const char *, struct stat *)=NULL;
int     (*__statptr)(const char *, struct stat *)=NULL;
#endif
int     (*__sscanfptr)(const char *, const char *, ...)=NULL;
char*   (*__fgetsptr)(char *, int , FILE *)=NULL;
FILE*   (*__fopenptr)(const char *, const char *)=NULL;
size_t  (*__strlenptr)(const char *)=NULL;
int     (*__readlinkptr)(const char * , char * , size_t )=NULL;
char*   (*__getloginptr)(void)=NULL;
void*   (*__memcpyptr)(void *, const void *, size_t)=NULL;
char*   (*__strncpyptr)(char *, const char *, size_t)=NULL;
void*   (*__mallocptr)(size_t)=NULL;
void*   (*__reallocptr)(void *, size_t)=NULL;
void    (*__freeptr)(void *)=NULL;
clock_t (*__timesptr)(struct tms *)=NULL;
int     (*__chmodptr)(const char *path, mode_t mode);
int     (*__chownptr)(const char *path, uid_t owner, gid_t group);

/*------------------------------------------------------------------------
 * trackle_funcptr_setup
 *
 * set up internal-use pointers to common libc functions to be used in
 * place of the actual functions
 *------------------------------------------------------------------------
 */
int trackle_funcptr_setup() {
    static int initialized=0;
    void *libc_ptr;

    /* don't run this twice */
    if (initialized)
        return 1;

    initialized++;

    /* open the library */
    libc_ptr = trackle_libopen(LIBC_PATH);

    /* look up things */
    __chmodptr = trackle_getfunc(libc_ptr,"chmod");
    __chownptr = trackle_getfunc(libc_ptr,"chown");
    __vprintfptr = trackle_getfunc(libc_ptr,"vprintf");
    __printfptr = trackle_getfunc(libc_ptr,"printf");
    __fprintfptr = trackle_getfunc(libc_ptr,"fprintf");
    __snprintfptr = trackle_getfunc(libc_ptr,"snprintf");
    __vsnprintfptr = trackle_getfunc(libc_ptr,"vsnprintf");
    __readptr = trackle_getfunc(libc_ptr,"read");
    __writeptr = trackle_getfunc(libc_ptr,"write");
    __fwriteptr = trackle_getfunc(libc_ptr,"fwrite");
    __openptr = trackle_getfunc(libc_ptr,"open");
    __closeptr = trackle_getfunc(libc_ptr,"close");
    __fcloseptr = trackle_getfunc(libc_ptr,"fclose");
#ifdef LINUX
    __statptr = trackle_getfunc(libc_ptr,"__xstat");
    __lstatptr = trackle_getfunc(libc_ptr,"__lxstat");
#else
    __statptr = trackle_getfunc(libc_ptr,"stat");
    __lstatptr = trackle_getfunc(libc_ptr,"lstat");
#endif
    __sscanfptr = trackle_getfunc(libc_ptr,"sscanf");
    __fopenptr = trackle_getfunc(libc_ptr,"fopen");
    __fgetsptr = trackle_getfunc(libc_ptr,"fgets");
    __strlenptr = trackle_getfunc(libc_ptr,"strlen");
    __readlinkptr = trackle_getfunc(libc_ptr,"readlink");
    __getloginptr = trackle_getfunc(libc_ptr,"getlogin");
    __memcpyptr = trackle_getfunc(libc_ptr,"memcpy");
    __strncpyptr = trackle_getfunc(libc_ptr,"strncpy");
    __mallocptr = trackle_getfunc(libc_ptr,"malloc");
    __reallocptr = trackle_getfunc(libc_ptr,"realloc");
    __freeptr = trackle_getfunc(libc_ptr,"free");
    __timesptr = trackle_getfunc(libc_ptr,"times");

    /* make sure that we got all the needed pointers */
    if (!(
        __readptr &&
        __writeptr &&
        __fwriteptr &&
        __openptr &&
        __vprintfptr &&
        __printfptr &&
        __fprintfptr &&
        __closeptr &&
        __fcloseptr &&
        __statptr &&
        __snprintfptr &&
        __sscanfptr &&
        __fopenptr &&
        __fgetsptr &&
        __lstatptr &&
        __readlinkptr &&
        __getloginptr &&
        __memcpyptr &&
        __strncpyptr &&
        __mallocptr &&
        __freeptr &&
        __timesptr &&
        __chmodptr &&
        __chownptr
    )) {
        fprintf(stderr, "common::Failed to find libc functions\n");
        exit(-1);
    }

    /* clean things up */
    dlclose(libc_ptr);

    return 0;
}


int trackle_stat(const char *path, struct stat *buf) {
#ifdef LINUX
    return ((*__statptr)(_STAT_VER, path, buf));
#else
    return( (*__statptr)(path,buf) );
#endif
}

// vim:sw=4
