/*
 * overrides.c
 *
 * This file contains the library methods which are overridden by
 * libtrackle. They are currently:
 *
 *    int   execv(const char *path, char *const argv[]);
 *    int   execvp(const char *file, char *const argv[]);
 *    int   open(const char *path, int oflag, mode_t mode);
 *    int   open64(const char *path, int oflag, mode_t mode);
 *    int   creat(const char *path, mode_t mode);
 *    int   chmod(const char *path, mode_t mode);
 *    int   chown(const char *path, uid_t owner, gid_t group);
 *    int   lchown(const char *path, uid_t owner, gid_t group);
 *    int   unlink(const char *path);
 *    int   truncate(const char *path, off_t length);
 *    FILE* fopen(const char *filename, const char *mode);
 *    int   symlink(const char *name1, const char *name2);
 *    int   rename(const char *old, const char *new);
 *    int   link(const char *existing, const char *new);
 *    int   mknod(const char *path, mode_t mode, dev_t dev);
 *
 *
 * 2006 Oct 5
 * Dan Crosta <dcrosta@sccs.swarthmore.edu>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>

#include "protos.h"
#include "paths.h"



int execv(const char *path, char *const argv[]) {
    static int (*fptr)(const char *, char *const argv[]) = NULL;

    TRACKLE_LOOKUP_COMMAND(int (*)(const char*, char* const argv[]), "execv", fptr, -1);

    trackle_fixenv();

    return (*fptr)(path, argv);
}

int execvp(const char *file, char *const argv[]) {
    static int (*fptr)(const char *, char *const argv[]) = NULL;

    TRACKLE_LOOKUP_COMMAND(int (*)(const char*, char* const argv[]), "execvp", fptr, -1);

    trackle_fixenv();

    return (*fptr)(file, argv);
}


int open(const char *path, int oflag, mode_t mode) {
    static  int (*fptr)(const char *, int , /* mode_t mode */...)=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, int , /* mode_t mode */...), "open",fptr,-1);

    trackle_logFileAccess(path);
    
    return((*fptr)(path,oflag, mode));
}

int open64(const char *path, int oflag, mode_t mode) {
    static  int (*fptr)(const char *, int , /* mode_t mode */...)=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, int , /* mode_t mode */...), "open64",fptr,-1);

    trackle_logFileAccess(path);
    
    return((*fptr)(path,oflag, mode));
}


int creat(const char *path, mode_t mode) {
    static  int (*fptr)(const char *, mode_t )=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, mode_t ),"creat",fptr,-1);

    trackle_logFileAccess(path);
    
    return((*fptr)(path,mode));
}


int chmod(const char *path, mode_t mode) {
    static  int (*fptr)(const char *, mode_t )=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, mode_t ),"chmod",fptr,-1);

    trackle_logFileAccess(path);

    return((*fptr)(path,mode));
}


int chown(const char *path, uid_t owner, gid_t group) {
    static  int (*fptr)(const char *, uid_t , gid_t )=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, uid_t , gid_t ),"chown",fptr, -1);

    trackle_logFileAccess(path);

    return((*fptr)(path,owner,group));
}


int lchown(const char *path, uid_t owner, gid_t group) {
    static  int (*fptr)(const char *, uid_t , gid_t )=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, uid_t , gid_t ),"lchown",fptr, -1);

    trackle_logFileAccess(path);

    return((*fptr)(path,owner,group));
}



int unlink(const char *path) {
    static  int (*fptr)(const char *)=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *),"unlink",fptr,-1);

    trackle_logFileAccess(path);

    return((*fptr)(path));
}


int truncate(const char *path, off_t length) {
    static  int (*fptr)(const char *, off_t )=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, off_t ),"truncate",fptr,-1);

    trackle_logFileAccess(path);

    return((*fptr)(path,length));
}


FILE* fopen(const char *filename, const char *mode) {
    static FILE *(*fptr)(const char *, const char *)=NULL;

    char exePath[PATH_MAX];
    char* exeName = NULL;
    int len;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( FILE *(*)(const char *, const char *),"fopen",fptr, NULL);

    /* lookup the currently executing program by
     * checking the dest of /proc/self/exe
     *
     * XXX: Linux only!
     */
    len = (*__readlinkptr)("/proc/self/exe", exePath, PATH_MAX);
    if ( errno ) {
        switch (errno) {
        case EINVAL:
            msg("/proc/self/exe is not a symbolic link.\n");
            break;
        case EIO:
            msg("encountered IO error tracing /proc/self/exe.\n");
            break;
        }

    } else {
        // readlink() does not NUL terminate
        exePath[len] = '\0';

        // get just the executable name (ie '/bin/bash' => 'bash')
        exeName = strrchr(exePath, '/');

        // ensure exeName works
        if ( NULL == exeName ) {
            exeName = exePath;
        } else {
            // clear the leading slash
            exeName++;
        }


        // log files if the accessing program is not sudo --
        // sudo clears the environment for a while during breaking
        // libtrackle until it creates a new environment for the
        // target program, which we fix in trackle_fixenv()
        if ( 0 != strcmp(exeName, "sudo") ) {
            trackle_logFileAccess(filename);
        }
    }

    return ((*fptr)(filename,mode));
}

int symlink(const char *name1, const char *name2) {
    static  int (*fptr)(const char *, const char *)=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, const char *),"symlink",fptr, -1);

    trackle_logFileAccess(name2);

    return((*fptr)(name1,name2));
}


int rename(const char *old, const char *new) {
    static  int (*fptr)(const char *, const char *)=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND( int (*)(const char *, const char *),"rename",fptr,-1);

    trackle_logFileAccess(old);
    trackle_logFileAccess(new);

    return((*fptr)(old,new));
}

int link(const char *existing, const char *new) {
    static int (*fptr)(const char *, const char *)=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND(int (*)(const char *, const char *),"link",fptr,-1);

    trackle_logFileAccess(new);

    return((*fptr)(existing,new));
}

int mknod(const char *path, mode_t mode, dev_t dev) {
    static int (*fptr)(const char *, mode_t , dev_t )=NULL;

    /* lookup next function */
    TRACKLE_LOOKUP_COMMAND(int (*)(const char *, mode_t , dev_t ),"mknod",fptr, -1);

    trackle_logFileAccess(path);

    return ((*fptr)(path,mode,dev));
}

// vim:sw=4
