import unittest
import os
import shutil
import random
import time
import stat
import pwd, grp
import types

def _p(*args):
    return os.path.join(os.getcwd(), *args)

def _f(err, msg):
    if msg != '':
        err = '%s [[%s]]' % (msg, err)
    return err

def statToDict(stat):
    """
    translates the contents of a .stat or .stat.end file
    to a python dict. takes lines of the form:

      '<key>: <value ..>'

    and creates dict entries <key> => <value>. assumes
    there are no duplicate keys in the stat
    """

    if type(stat) in types.StringTypes:
        stat = stat.splitlines()
    
    out = {}
    for line in stat:
        line = line.strip()
        try:
            (key, val) = line.split(':')
            val = val.strip()
            out[key] = val
        except ValueError:
            if len(line) and line[-1] == ':':
                out[line[:-1]] = ''

            elif len(line):
                out[line] = ''
    
    return out


class TrackleTestCase(unittest.TestCase):
    """
    specialized TestCase with extra assertion methods
    that check things about the session state dir created
    by libtrackle
    """

    session_parent_dir = 'sessions'
    trackleInitialized = False


    # these are automatically removed by
    # parent's cleanUp() method
    _output = []

    # TrackleTestCase's setUp method will create
    # files with the given (name, contents, extra) for
    # each sequence in _context. extra is a list of
    # 2-tuples in the form (cmd, val).
    #
    # cmd is one of:
    #   mode - sets the mode to val (integer)
    #   link - a symbolic link named val is created pointing to the created file
    #   owner - val is (uid, gid): the file's ownership is set
    _context = []


    def _file(self, *args):
        return os.path.join( self.filesDir, *args )


    def setUp(self):
        self.setupSessionParentDir()

        for filenem in self._output:
            try:
                os.remove(filename)
            except:
                pass
    
        for item in self._context:
            extra = []

            if len(item) == 2:
                name, contents = item

            elif len(item) == 3:
                name, contents, extra = item
            
            fp = file(name, 'w')
            fp.write(contents)
            fp.close()

            for (cmd, arg) in extra:
                if cmd == 'mode':
                    os.spawnvp(os.P_WAIT, 'sudo', ['sudo', 'chmod', '%o' % arg, name])

                elif cmd == 'link':
                    os.symlink(name, arg)

                elif cmd == 'owner':
                    os.spawnvp(os.P_WAIT, 'sudo', ['sudo', 'chown', '%s:%s' % arg, name])




        self.startTrackle()

    def tearDown(self):
        self.cleanupSessionParentDir()

        for filename in self._output:
            try:
                os.spawnvp(os.P_WAIT, 'sudo', ['sudo', 'rm', '-f', filename])
            except:
                pass
    

    def setupSessionParentDir(self):
        """
        sets up a directory in which to create session
        directories... confusing, i know, but it gives
        us a convenient place to do all our cleanup
        """

        try:
            os.makedirs( self.session_parent_dir )
        except:
            pass


    def cleanupSessionParentDir(self):
        """
        deletes the session parent directory and everything
        inside it
        """

        shutil.rmtree( self.session_parent_dir )


    def runPlain(self, *args):
        """
        runs a system program by itself
        each command line token goes as a separate argument,
        ie: runClean('cat', '/path/to/file')
        """

        if len(args) == 0:
            return None

        ret = os.spawnve(os.P_WAIT, args[0], args, self.trackleEnv)

        self._updateIndexCache()

        return ret


    def runPlainPath(self, *args):
        """
        runs a system program by itself
        each command line token goes as a separate argument,
        ie: runClean('cat', '/path/to/file')
        """

        if len(args) == 0:
            return None

        ret = os.spawnvpe(os.P_WAIT, args[0], args, self.trackleEnv)

        self._updateIndexCache()

        return ret


    def runSudo(self, *args):
        """
        runs a system program as the current user with a call
        like "sudo -u <thisuser> ..."
        """

        if len(args) == 0:
            return None

        # munge the first argument to a full path
        # relative to the current directory
        args = list(args)
        args[0] = os.path.join( os.getcwd(), args[0] )

        myargs = ['sudo']
        myargs.extend(args)

        ret = os.spawnvpe(os.P_WAIT, 'sudo', myargs, self.trackleEnv)

        self._updateIndexCache()

        return ret


    def _updateIndexCache(self):
        """
        updates indexLines
        """

        fp = file(os.path.join(self.sessionDir, 'index'), 'r')
        self.indexLines = [line.strip() for line in fp]
        fp.close()
    

    def startTrackle(self):
        """
        begins trackle monitoring for any processes spawned
        as a child of this process. sets up environment
        variables, session dir, etc.
        """

        if not self.trackleInitialized:
            self.ticketId = random.randint(0, 100)
            self.startTime = time.time()
            self.sessionID = '%d-%d' % (self.ticketId, self.startTime)

            self.sessionDir = os.path.join(self.session_parent_dir, self.sessionID)
            self.sessionDir = os.path.normpath(self.sessionDir)
            self.filesDir   = os.path.join(self.sessionDir, 'files')

            os.makedirs(self.sessionDir)
            os.makedirs(self.filesDir)

            
            # create the intial trackle environment by
            # copying the current env
            self.trackleEnv = {}
            for key in os.environ.keys():
                self.trackleEnv[key] = os.environ[key]


            # get the date as returned by `date`
            self.startDate = time.strftime('%a %b %d %X %Z %Y', time.localtime(self.startTime))

            self.trackleEnv['SHELL'] = '/bin/bash'
            os.mknod(os.path.join(self.sessionDir, 'hist.bash'))
            if 'HISTFILE' in self.trackleEnv:
                self.oldhistfile = self.trackleEnv['HISTFILE']
            self.trackleEnv['HISTFILE'] = os.path.join(self.sessionDir, 'hist.bash')

            self.trackleEnv['TRACKLE_SESSION_DIR'] = self.sessionDir
            self.trackleEnv['TRACKLE_FILE_COUNTER'] = os.path.join(self.sessionDir, 'counter')
            self.trackleEnv['TRACKLE_UID'] = str(os.getuid())

            self.counterFileName = os.path.join(self.sessionDir, 'counter')
            self.indexFileName = os.path.join(self.sessionDir, 'index')
            os.mknod(self.counterFileName)
            os.mknod(self.indexFileName)

            self.trackleEnv['LD_PRELOAD'] = 'libtrackle.so'


        self.trackleInitialized = True


    def stopTrackle(self):
        """
        clean up local state after a Trackle session
        """

        del self.trackleEnv

        del self.sessionDir
        del self.ticketId
        del self.startTime
        del self.filesDir

        del self.indexLines
        del self.indexNames
        del self.indexNums

        self.trackleInitialized = False


    def _lookupFilenum(self, filename):
        """
        return the file number of the given filename
        """

        filenum = None
        for line in self.indexLines:
            (name, num) = line.split(',')

            if name == filename:
                filenum = num

        if filenum is None:
            raise self.failureException, "Could not find '%s' in index" % filename

        return filenum

    def assertStatsAndFilesCorrect(self, msg = ''):
        """
        verifies that for each file in the index:

          - there is a .stat and a .stat.end
            - each of the above contains at least the 'special' field
            - each of the above is either a complete stat dump or it has 'noexist' in special
            - if 'noexist' or 'unreadable(uid)' not in special, then the .file or .file.end
              exists respectively
        """

        # all the stat fields we expect in a full dump
        statFields = ['dev', 'inode', 'mode', 'nlink', 'uid', 'gid', 'rdev', 'size', 'blksize', 'blocks', 'mtime', 'ctime']

        for line in self.indexLines:
            (name, num) = line.split(',')

            self.assert_(os.path.exists(self._file('%s.stat' % num)), _f("missing .stat for '%s'" % name, msg))
            self.assert_(os.path.exists(self._file('%s.stat.end' % num)), _f("missing .stat.end for '%s'" % name, msg))

            fp = file(self._file('%s.stat' % num), 'r')
            contents = fp.readlines()
            fp.close()
            sstat = statToDict(contents)

            fp = file(self._file('%s.stat.end' % num), 'r')
            contents = fp.readlines()
            fp.close()
            fstat = statToDict(contents)

            self.assert_('special' in sstat, _f("'special' missing from .stat for '%s'" % name, msg))
            self.assert_('special' in fstat, _f("'special' missing from .stat.end for '%s'" % name, msg))
            
            if 'noexist' in sstat['special']:
                self.assert_(not os.path.exists(self._file('%s.file' % num)), _f(".file exists for '%s' when 'noexist' in special" % name, msg))

            else:
                if 'unreadable' in sstat['special']:
                    self.assert_(not os.path.exists(self._file('%s.file' % num)), _f(".file exists for '%s' when 'unreadable' in special" % name, msg))
                else:
                    self.assert_(os.path.exists(self._file('%s.file' % num)), _f(".file does not exist for '%s'" % name, msg))

                for field in statFields:
                    self.assert_(field in sstat, _f("missing field '%s' in .stat for '%s'" % (field, name), msg))

            
            if 'noexist' in fstat['special']:
                self.assert_(not os.path.exists(self._file('%s.file.end' % num)), _f(".file exists for '%s' when 'noexist' in special" % name, msg))

            else:
                if 'unreadable' in fstat['special']:
                    self.assert_(not os.path.exists(self._file('%s.file.end' % num)), _f(".file.end exists for '%s' when 'unreadable' in special" % name, msg))
                else:
                    self.assert_(os.path.exists(self._file('%s.file.end' % num)), _f(".file.end does not exist for '%s'" % name, msg))

                for field in statFields:
                    self.assert_(field in fstat, _f("missing field '%s' in .stat.end for '%s'" % (field, name), msg))


            
            



    def assertIndexFileCorrect(self, files = [], strict = False, msg = ''):
        """
        verifies that the index file was correctly created
        and updated for the session, and that it contains
        all and only the files listed in the `files` param

        fails if:
          - any file listed in `files` param is in the index
          - all index file numbers are unique
          - all files in the index are in `files` (iff strict == True)
        """

        self.indexNums = {}
        self.indexNames = {}

        for line in self.indexLines:
            (name, num) = line.split(',')

            if num in self.indexNums:
                raise self.failureException, _f("Duplicate index number '%s' in index" % num, msg)
            self.indexNums[num] = line

            if name in self.indexNames:
                raise self.failrueException, _f("File in index more than once: '%s'" % name, msg)
            self.indexNames[name] = file

            if strict and name not in files:
                raise self.failureException, _f("Extra file in index: '%s'" % name, msg)


        for name in files:
            if name not in self.indexNames:
                raise self.failureException, _f("File '%s' not in index" % name, msg)


    def assertIndexFileConsecutive(self, msg = ''):
        """
        fail if there are gaps in the numbers in the index
        file, if the numbers appear out of order, or if
        the final number is not the same as number of lines
        in the counter file (this is how the counter file
        currently tracks the number of files it knows about)
        """

        ctr = 0

        for line in self.indexLines:
            (name, num) = line.split(',')

            num = int(num)
            if num != ctr:
                raise self.failureException, _f("File index for '%s' was '%d' instead of expected '%d'" % (name, num, ctr), msg)

            ctr += 1

        fp = file(os.path.join(self.sessionDir, 'counter'), 'r')
        counter = len(fp.readlines())
        fp.close()

        if counter != ctr:
            raise self.failureException, _f("Counter file says %d, last index is %d" % (counter, ctr), msg)


    def assertIndexHas(self, filename, msg = ''):
        """
        fail if the index does not have the given file
        """

        found = False

        for line in self.indexLines:
            (name, num) = line.split(',')

            if name == filename:
                found = True
                break

        if not found:
            raise self.failureException, _f("File '%s' not in index" % filename, msg)

    
    def assertIndexDoesNotHave(self, filename, msg = ''):
        """
        fail if the index does have the given file
        """

        found = False

        for line in self.indexLines:
            (name, num) = line.split(',')

            if name == filename:
                found = True
                break

        if found:
            raise self.failureException, _f("Invalid file '%s' in index" % filename, msg)


    def assertWasCreated(self, filename, msg = ''):
        """
        checks that the .file and .stat files indicate
        that the given file was created, and that it
        existed at the end of the session. more precisely:

         - <num>.file should not exist
         - <num>.file.end should exist
         - <num>.stat should have special value 'noexist'
         - <num>.stat.end should not have special value 'noexist'

        Note that <num>.file.end may be correctly empty, so
        we do not check that it has any contents.
        """

        filenum = self._lookupFilenum(filename)

        
        if os.path.exists( self._file('%s.file' % filenum) ):
            raise self.failureException, _f("'%s.file' existed (for created file '%s')" % (filenum, filename), msg)


        special = self._getStatFileField( self._file('%s.stat' % filenum), 'special' )
        if 'noexist' not in special:
            raise self.failureException, _f("'%s.stat' did not contain 'noexist' (for created file '%s')" % (filenum, filename), msg)


        if not os.path.exists( self._file('%s.file.end' % filenum) ):
            raise self.failureException, _f("'%s.file.end' did not exist (for created file '%s')" % (filenum, filename), msg)


        special = self._getStatFileField( self._file('%s.stat.end' % filenum), 'special' )
        if 'noexist' in special:
            raise self.failureException, _f("'%s.stat.end' contained 'noexist' (for created file '%s')" % (filenum, filename), msg)


    def assertWasNotCreated(self, filename, msg = ''):
        """
        checks that the .file and .stat files indicate
        that this file existed before the session started.
        more precisely:

         - <num>.stat should not contain special value 'noexist'
        """

        filenum = self._lookupFilenum(filename)

        special = self._getStatFileField( self._file('%s.stat' % filenum), 'special' )
        if 'noexist' in special:
            raise self.failureException, _f("'%s.stat' contained 'noexist' (for file '%s')" % (filenum, filename), msg)



    def assertWasDeleted(self, filename, msg = ''):
        """
        checks that the .file.end and .stat.end files
        indicate that this file was deleted during
        the session. more precisely:

         - <num>.file should exist
         - <num>.file.end should not exist
         - <num>.stat should not contain special value 'noexist'
         - <num>.stat.end should contain special value 'noexist' 
        """

        filenum = self._lookupFilenum(filename)

        
        if not os.path.exists( self._file('%s.file' % filenum) ):
            raise self.failureException, _f("'%s.file' did not exist (for deleted file '%s')" % (filenum, filename), msg)


        special = self._getStatFileField( self._file('%s.stat' % filenum), 'special' )
        if 'noexist' in special:
            raise self.failureException, _f("'%s.stat' contained 'noexist' (for deleted file '%s')" % (filenum, filename), msg)


        if os.path.exists( self._file('%s.file.end' % filenum) ):
            raise self.failureException, _f("'%s.file.end' existed (for deleted file '%s')" % (filenum, filename), msg)


        special = self._getStatFileField( self._file('%s.stat.end' % filenum), 'special' )
        if 'noexist' not in special:
            raise self.failureException, _f("'%s.stat.end' did not contain 'noexist' (for deleted file '%s')" % (filenum, filename), msg)



    def assertWasNotDeleted(self, filename, msg = ''):
        """
        checks that the .file.end and .stat.end
        files indicate the file existed at the end
        of the session. specifically:

         - <num>.file.end should exist
         - <num>.stat.end shoud not contain special value 'noexist'
        """

        filenum = self._lookupFilenum(filename)

        if not os.path.exists( self._file('%s.file.end' % filenum) ):
            raise self.failureException, _f("'%s.file.end' did not exist (for not-deleted file '%s')" % (filenum, filename), msg)

        special = self._getStatFileField( self._file('%s.stat.end' % filenum), 'special' )
        if 'noexist' in special:
            raise self.failureExcepton, _f("'%s.stat.end' contained 'noexist' (for not-deleted file '%s')" % (filenum, filename), msg)


    def assertWasNotModified(self, filename, msg = ''):
        """
        checks the mtime field of the .stat and .stat.end,
        and fails if they are not identical
        """

        filenum = self._lookupFilenum(filename)

        smtime = self._getStatFileField(os.path.join(self.filesDir, '%s.stat' % filenum), 'mtime')
        emtime = self._getStatFileField(os.path.join(self.filesDir, '%s.stat.end' % filenum), 'mtime')

        if smtime != emtime:
            raise self.faiulreException, _f("File `%s' was modified (mtimes do not agree)" % filename, msg)


    def assertBeforeAndAfterContents(self, filename, before, after, msg = ''):
        """
        checks that the given filename has the exact
        same contents as the strings `before' and `after'.
        specifying None for either before or after
        will skip the check for that state
        """

        filenum = self._lookupFilenum(filename)

        if before is not None:
            try:
                fp = file(self._file('%s.file' % filenum))
                contents = fp.read()
                fp.close()
            except IOError, e:
                if e.errno == 2:
                    raise self.failureException, _f("Pre-Trackle contents file '%s.file' (for '%s') missing" % (filenum, filename), msg)


            if contents != before:
                raise self.failureException, _f("Pre-Trackle contents of '%s' ('%s.file') not correct. Should be:\n`%s`\n\nWas:\n`%s`" % (filename, filenum, before, contents), msg)


        if after is not None:
            try:
                fp = file(self._file('%s.file.end' % filenum))
                contents = fp.read()
                fp.close()
            except IOError, e:
                if e.errno == 2:
                    raise self.failureException, _f("Post-Trackle contents file '%s.file.end' (for '%s') missing" % (filenum, filename), msg)

            if contents != after:
                raise self.failureException, _f("Post-Trackle contents of '%s' ('%s.file.end') not correct. Should be:\n`%s`\n\nWas:\n`%s`" % (filename, filenum, after, contents), msg)


    
    def _getStatFileField(self, realfilename, field):
        """
        get a field from a .stat or .stat.end file
        """
        
        field += ':'

        fp = file(realfilename, 'r')
        
        for line in fp:
            (tok, val) = line.split(' ', 1)
            if tok == field:
                return val


    def _statFileHasPermissions(self, realfilename, mode):

        actual = int(self._getStatFileField(realfilename, 'mode'), 8)

        if stat.S_IMODE(actual) == int(mode, 8):
            return True

        return False
                
        
    def assertFileHasPreTracklePermissions(self, filename, mode, msg = ''):
        """
        fails if the file had pre-Trackle permissions
        other than that specified in mode. mode specified
        as octal string.
        """

        filenum = self._lookupFilenum(filename)
        realfilename = os.path.join(self.filesDir, '%s.stat' % filenum)

        if not self._statFileHasPermissions(realfilename, mode):
            actual = int(self._getStatFileField(realfilename, 'mode'), 8)
            actual = stat.S_IMODE(actual)
            actual = '%o' % actual
            raise self.failureException, _f("File `%s' has wrong pre-trackle permissions: had %s instead of %s" % (filename, actual, mode), msg)
    

    def assertFileHasPostTracklePermissions(self, filename, mode, msg = ''):
        """
        fails if the file had post-Trackle permissions
        other than that specified in mode. mode specified
        as octal string.
        """

        filenum = self._lookupFilenum(filename)
        realfilename = os.path.join(self.filesDir, '%s.stat.end' % filenum)

        if not self._statFileHasPermissions(realfilename, mode):
            actual = int(self._getStatFileField(realfilename, 'mode'), 8)
            actual = stat.S_IMODE(actual)
            actual = '%o' % actual
            raise self.failureException, _f("File `%s' has wrong post-trackle permissions: had %s instead of %s" % (filename, actual, mode), msg)


    def assertFileHasPreTrackleOwner(self, filename, owner, msg = ''):
        """
        fail if the file has pre-Trackle owner other than
        the named owner. owner may be specified as string
        username or integer uid
        """
        filenum = self._lookupFilenum(filename)
        realfilename = os.path.join(self.filesDir, '%s.stat' % filenum)

        uid = int(self._getStatFileField(realfilename, 'uid'))
        passwd = pwd.getpwuid(uid)
        uname = passwd.pw_name

        if type(owner) in types.StringTypes:
            if not uname == owner:
                raise self.failureException, _f("File has wrong pre-trackle owner: %s instead of %s" % (uname, owner), msg)

        else:
            if not uid == owner:
                raise self.failureException, _f("File has wrong pre-trackle owner: %s instead of %s" % (uname, owner), msg)


    def assertFileHasPostTrackleOwner(self, filename, owner, msg = ''):
        """
        fail if the file has post-Trackle owner other than
        the named owner. owner may be specified as string
        username or integer uid
        """
        filenum = self._lookupFilenum(filename)
        realfilename = os.path.join(self.filesDir, '%s.stat.end' % filenum)

        uid = int(self._getStatFileField(realfilename, 'uid'))
        passwd = pwd.getpwuid(uid)
        uname = passwd.pw_name

        if type(owner) in types.StringTypes:
            if not uname == owner:
                raise self.failureException, _f("File has wrong post-trackle owner: %s instead of %s" % (uname, owner), msg)

        else:
            if not uid == owner:
                raise self.failureException, _f("File has wrong post-trackle owner: %s instead of %s" % (uname, owner), msg)
    

################
## Test Cases ##
################



class GeneralTestCase(TrackleTestCase):
    """
    tests general libtrackle behavior using fopen
    """

    _output  = ['general.out', 'general2.out', 'general3.out']
    _context = [
        ('general2.out', 'Pre-start contents\n'),
        ('general3.out', 'the file contents\n', [('owner', ('root', 'root')),
                                                 ('mode',  600)])
        ]

    def test_general_IndexFileCorrect(self):
        self.runPlain('my-fopen', 'general.out', 'this is a string to append')
        self.runPlain('my-fopen', 'general2.out', 'this is a string to append')

        self.assertIndexFileCorrect()
        self.assertStatsAndFilesCorrect()


    def test_general_IndexCounterConsecutive(self):
        self.runPlain('my-fopen', 'general.out', 'this is a string to append')
        self.runPlain('my-fopen', 'general2.out', 'this is a string to append')

        self.assertIndexFileConsecutive()


    def test_general_IndexHasFiles(self):
        self.runPlain('my-fopen', 'general.out', 'this is a string to append')
        self.runPlain('my-fopen', 'general2.out', 'this is a string to append')

        self.assertIndexHas(_p('general.out'))
        self.assertIndexHas(_p('general2.out'))

        self.assertIndexFileCorrect(
                [_p('general.out'),
                 _p('general2.out')],
                strict = True)


    def test_general_Creation(self):
        self.runPlain('my-fopen', 'general.out', 'this is a string to append')
        self.runPlain('my-fopen', 'general2.out', 'this is a string to append')

        self.assertWasCreated(_p('general.out'))
        self.assertWasNotCreated(_p('general2.out'))


    def test_general_SudoIndexFileCorrect(self):
        self.runSudo('my-fopen', 'general.out', 'this is a string to append')
        self.runSudo('my-fopen', 'general2.out', 'this is a string to append')

        self.assertIndexFileCorrect()


    def test_general_SudoIndexCounterConsecutive(self):
        self.runSudo('my-fopen', 'general.out', 'this is a string to append')
        self.runSudo('my-fopen', 'general2.out', 'this is a string to append')

        self.assertIndexFileConsecutive()


    def test_general_SudoIndexHasFiles(self):
        self.runSudo('my-fopen', 'general.out', 'this is a string to append')
        self.runSudo('my-fopen', 'general2.out', 'this is a string to append')

        self.assertIndexHas(_p('general.out'))
        self.assertIndexHas(_p('general2.out'))

        self.assertIndexFileCorrect(
                [_p('general.out'),
                 _p('general2.out')],
                strict = True)


    def test_general_SudoCreation(self):
        self.runSudo('my-fopen', 'general.out', 'this is a string to append')
        self.runSudo('my-fopen', 'general2.out', 'this is a string to append')

        self.assertWasCreated(_p('general.out'))
        self.assertWasNotCreated(_p('general2.out'))


    def test_general_SudoReadThenPlainRead(self):
        self.runSudo('my-fopen', 'general3.out')
        self.runPlain('my-fopen', 'general3.out')

        self.assertWasNotCreated(_p('general3.out'))
        self.assertWasNotDeleted(_p('general3.out'))
        self.assertWasNotModified(_p('general3.out'))
        self.assertBeforeAndAfterContents(_p('general3.out'),
                before = 'the file contents\n',
                after  = 'the file contents\n',
                msg = "The file contents appear to have been deleted")


    def test_general_PlainReadThenSudoRead(self):
        self.runPlain('my-fopen', 'general3.out')
        self.runSudo('my-fopen', 'general3.out')

        self.assertWasNotCreated(_p('general3.out'))
        self.assertWasNotDeleted(_p('general3.out'))
        self.assertWasNotModified(_p('general3.out'))
        self.assertBeforeAndAfterContents(_p('general3.out'),
                before = 'the file contents\n',
                after  = 'the file contents\n',
                msg = "The file contents appear to have been added")




class FopenTestCase(TrackleTestCase):

    _output  = ['fopen.out', 'fopen2.out']
    _context = [('fopen2.out', 'This is the pre-run context.\n')]
    

    def test_fopen_Creation(self):
        self.runPlain('my-fopen', 'fopen.out', 'creating a file with this string')
        
        self.assertWasCreated(_p('fopen.out'))

    def test_fopen_Appending(self):
        self.runPlain('my-fopen', 'fopen2.out', 'appending this string to a file')

        self.assertIndexHas(_p('fopen2.out'))
        self.assertWasNotCreated(_p('fopen2.out'))

    def test_fopen_Correct(self):
        self.runPlain('my-fopen', 'fopen.out', 'The sole contents of\nthis file.')

        self.assertBeforeAndAfterContents(
                _p('fopen.out'),
                None,
                'The sole contents of\nthis file.\n')

        self.runPlain('my-fopen', 'fopen2.out', 'The additional contents of\nthis file.')

        self.assertBeforeAndAfterContents(
                _p('fopen2.out'),
                'This is the pre-run context.\n',
                'This is the pre-run context.\nThe additional contents of\nthis file.\n')



class OpenTestCase(TrackleTestCase):

    _output  = ['open.out', 'open2.out']
    _context = [('open2.out', 'This is the pre-run context.\n')]
    

    def test_open_Creation(self):
        self.runPlain('my-open', 'open.out', 'creating a file with this string')
        
        self.assertWasCreated(_p('open.out'))

    def test_open_Appending(self):
        self.runPlain('my-open', 'open2.out', 'appending this string to a file')

        self.assertIndexHas(_p('open2.out'))
        self.assertWasNotCreated(_p('open2.out'))

    def test_open_Correct(self):
        self.runPlain('my-open', 'open.out', 'The sole contents of\nthis file.')

        self.assertBeforeAndAfterContents(
                _p('open.out'),
                None,
                'The sole contents of\nthis file.\n')

        self.runPlain('my-open', 'open2.out', 'The additional contents of\nthis file.')

        self.assertBeforeAndAfterContents(
                _p('open2.out'),
                'This is the pre-run context.\n',
                'This is the pre-run context.\nThe additional contents of\nthis file.\n')



class Open64TestCase(TrackleTestCase):

    _output  = ['open64.out', 'open64-2.out']
    _context = [('open64-2.out', 'This is the pre-run context.\n')]
    

    def test_open64_Creation(self):
        self.runPlain('my-open64', 'open64.out', 'creating a file with this string')
        
        self.assertWasCreated(_p('open64.out'))

    def test_open64_Appending(self):
        self.runPlain('my-open64', 'open64-2.out', 'appending this string to a file')

        self.assertIndexHas(_p('open64-2.out'))
        self.assertWasNotCreated(_p('open64-2.out'))

    def test_open64_Correct(self):
        self.runPlain('my-open64', 'open64.out', 'The sole contents of\nthis file.')

        self.assertBeforeAndAfterContents(
                _p('open64.out'),
                None,
                'The sole contents of\nthis file.\n')

        self.runPlain('my-open64', 'open64-2.out', 'The additional contents of\nthis file.')

        self.assertBeforeAndAfterContents(
                _p('open64-2.out'),
                'This is the pre-run context.\n',
                'This is the pre-run context.\nThe additional contents of\nthis file.\n')

class CreatTestCase(TrackleTestCase):

    _output = ['creat.out', 'creat2.out']

    def test_creat_Creation(self):
        self.runPlain('my-creat', 'creat.out')

        self.assertWasCreated(_p('creat.out'))

    def test_creat_Permission(self):
        self.runPlain('my-creat', 'creat.out')
        self.assertFileHasPostTracklePermissions(_p('creat.out'), '644')

        self.runPlain('my-creat', 'creat2.out', '775')
        self.assertFileHasPostTracklePermissions(_p('creat2.out'), '775')


class ChmodTestCase(TrackleTestCase):

    _output = ['chmod.out', 'chmod-zero.out']
    _context = [('chmod.out', ' ', [('mode', 0644)]),
                ('chmod-zero.out', ' ',  [('mode', 0644)])]

    def test_chmod_ZeroPermissions(self):
        self.runPlain('my-chmod', 'chmod-zero.out', '000')

        self.assertFileHasPostTracklePermissions(_p('chmod-zero.out'), '000')

    def test_chmod_Unreadable(self):
        self.runPlain('my-chmod', 'chmod.out', '200')

        self.assertFileHasPostTracklePermissions(_p('chmod.out'), '200')

    def test_chmod_PermissionChanges(self):
        self.runPlain('my-chmod', 'chmod.out', '0600')

        self.assertFileHasPreTracklePermissions(_p('chmod.out'), '644')
        self.assertFileHasPostTracklePermissions(_p('chmod.out'), '600')

    def test_chmod_ContentsAfterUnreadable(self):
        self.runPlain('my-chmod', 'chmod.out', '200')

        self.assertFileHasPostTracklePermissions(_p('chmod.out'), '200')
        self.assertBeforeAndAfterContents(_p('chmod.out'), before = None, after = ' ')

    def test_chmod_ContentsAfterZeroPermissions(self):
        self.runPlain('my-chmod', 'chmod.out', '000')

        self.assertFileHasPostTracklePermissions(_p('chmod-zero.out'), '000')
        self.assertBeforeAndAfterContents(_p('chmod.out'), before = None, after = ' ')

    def test_chmod_ContentsAfterPermissionChanges(self):
        self.runPlain('my-chmod', 'chmod.out', '0600')

        self.assertBeforeAndAfterContents(_p('chmod.out'), before = None, after = ' ')


class ChownTestCase(TrackleTestCase):

    _output = ['chown.out']
    _context = [('chown.out', ' ', [('mode', 0644)])]

    def test_chown_NoChangePermissions(self):
        self.runPlain('my-chown', 'chown.out', 'www-data')
        
        self.assertFileHasPostTracklePermissions(_p('chown.out'), '0644')


    def test_chown_SudoNoChangePermissions(self):
        self.runSudo('my-chown', 'chown.out', 'www-data')
        
        self.assertFileHasPostTracklePermissions(_p('chown.out'), '0644')


    def test_chown_ChangeOwner(self):
        self.runPlain('my-chown', 'chown.out', 'www-data')
        
        self.assertFileHasPreTrackleOwner(_p('chown.out'), os.getlogin())
        self.assertFileHasPostTrackleOwner(_p('chown.out'), os.getlogin())


    def test_chown_SudoChangeOwner(self):
        self.runSudo('my-chown', 'chown.out', 'www-data')
        
        self.assertFileHasPreTrackleOwner(_p('chown.out'), os.getlogin())
        self.assertFileHasPostTrackleOwner(_p('chown.out'), 'www-data')


class LchownTestCase(TrackleTestCase):

    _output = ['lchown.out', 'link-lchown.out']
    _context = [('lchown.out', ' ', [('link', 'link-lchown.out')])]

    def test_lchown_NoChangePermissions(self):
        self.runPlain('my-lchown', 'link-lchown.out', 'www-data')
        
        self.assertFileHasPostTracklePermissions(_p('link-lchown.out'), '0644')


    def test_lchown_SudoNoChangePermissions(self):
        self.runSudo('my-lchown', 'link-lchown.out', 'www-data')
        
        self.assertFileHasPostTracklePermissions(_p('link-lchown.out'), '0644')


    def test_lchown_ChangeOwner(self):
        self.runPlain('my-lchown', 'link-lchown.out', 'www-data')
        
        self.assertFileHasPreTrackleOwner(_p('link-lchown.out'), os.getlogin())
        self.assertFileHasPostTrackleOwner(_p('link-lchown.out'), os.getlogin())


    def test_lchown_SudoChangeOwner(self):
        self.runSudo('my-lchown', 'link-lchown.out', 'www-data')
        
        self.assertFileHasPreTrackleOwner(_p('link-lchown.out'), os.getlogin())
        self.assertFileHasPostTrackleOwner(_p('link-lchown.out'), 'www-data', "libtrackle needs to use lstat()?")


class UnlinkTestCase(TrackleTestCase):


    _context = [('unlink.in', 'blah\n')]


    def test_unlink_Working(self):
        self.runPlain('my-unlink', 'unlink.in')

        self.assertWasDeleted(_p('unlink.in'))


class TruncateTestCase(TrackleTestCase):

    _output = ['truncate.out']
    _context = [('truncate.out', '1234567890123456789\n')]


    def test_truncate_Shorten(self):
        self.runPlain('my-truncate', 'truncate.out', '10')

        self.assertBeforeAndAfterContents(_p('truncate.out'),
                before = '1234567890123456789\n',
                after = '1234567890'
                )


    def test_truncate_Lengthen(self):
        self.runPlain('my-truncate', 'truncate.out', '22')

        self.assertBeforeAndAfterContents(_p('truncate.out'),
                before = '1234567890123456789\n',
                after ='1234567890123456789\n\0\0' 
                )


class SymlinkTestCase(TrackleTestCase):

    _output = ['symlink.in', 'symlink.out']
    _context = [('symlink.in', 'hello, world\n')]


    def test_symlink_Created(self):
        self.runPlain('my-symlink', 'symlink.in', 'symlink.out')

        self.assertWasCreated(_p('symlink.out'))


    def test_symlink_CorrectAfterContents(self):
        self.runPlain('my-symlink', 'symlink.in', 'symlink.out')

        self.assertBeforeAndAfterContents(_p('symlink.out'),
                after = 'hello, world\n',
                before = None)


class RenameTestCase(TrackleTestCase):

    _output = ['rename.out']
    _context = [('rename.in', 'hello, world\n')]


    def test_rename_Moved(self):
        self.runPlain('my-rename', 'rename.in', 'rename.out')

        self.assertWasCreated(_p('rename.out'))
        self.assertWasDeleted(_p('rename.in'))


    def test_rename_CorrectAfterContents(self):
        self.runPlain('my-rename', 'rename.in', 'rename.out')

        self.assertBeforeAndAfterContents(_p('rename.out'),
                after = 'hello, world\n',
                before = None)
        self.assertBeforeAndAfterContents(_p('rename.in'),
                before = 'hello, world\n',
                after = None)


class LinkTestCase(TrackleTestCase):

    _output = ['link.in', 'link.out']
    _context = [('link.in', 'hello, world\n')]


    def test_link_Created(self):
        self.runPlain('my-link', 'link.in', 'link.out')

        self.assertWasCreated(_p('link.out'))


    def test_link_CorrectAfterContents(self):
        self.runPlain('my-link', 'link.in', 'link.out')

        self.assertBeforeAndAfterContents(_p('link.out'),
                after = 'hello, world\n',
                before = None)


class MknodTestCase(TrackleTestCase):

    _output = ['mknod.out']

    def test_mknod_Created(self):
        ret = self.runPlain('my-mknod', 'mknod.out', '0644')
        self.assert_(ret == 0)

        self.assertWasCreated(_p('mknod.out'))


    def test_mknod_Permissions(self):
        self.runPlain('my-mknod', 'mknod.out', '0644')

        self.assertFileHasPostTracklePermissions(_p('mknod.out'), '644')
