#!/usr/bin/env python

import os
import os.path
import stat
import sys
import string
from glob import glob
from distutils.core import setup
from distutils.command.install import install
from distutils.command.install_data import install_data
from distutils.command.install_scripts import install_scripts
from distutils.command.build import build
from distutils.command.clean import clean
from stat import ST_MODE, S_ISDIR

import trackle

PACKAGE = 'Trackle'
VERSION = str(trackle.__version__)
URL = trackle.__url__
LICENSE = trackle.__license__

if sys.version_info < (2, 3):
    print >>sys.stderr, 'You need at least Python 2.3 for %s %s' \
                        % (PACKAGE, VERSION)
    sys.exit(3)

def _p(unix_path):
     return os.path.normpath(unix_path)

class my_build(build):

    def run(self):
        # first run parent's build
        build.run(self)

        # calculate the path to setup.py
        my_path = _p(os.path.join(os.getcwd(), __file__))
        my_path = os.path.dirname(my_path)

        interpose_dir = _p(os.path.join(my_path, 'interpose'))

        # now build libtrackle.so
        self.spawn(('make', '-C', '%(interpose_dir)s' %
                {'interpose_dir': interpose_dir}))

        self.mkpath(_p(os.path.join(my_path, 'build', 'libs')))
        self.copy_file(
                _p(os.path.join(interpose_dir, 'libtrackle.so')),
                _p(os.path.join(my_path, 'build', 'libs', 'libtrackle.so')),
                )


class my_clean(clean):

    def run(self):
        import shutil

        # first run parent's build
        clean.run(self)

        # now clean the interpose/ directory
        my_path = _p(os.path.join(os.getcwd(), __file__))
        my_path = os.path.dirname(my_path)
        interpose_dir = _p(os.path.join(my_path, 'interpose'))

        self.spawn(('make', '-C', '%(interpose_dir)s' %
                {'interpose_dir': interpose_dir}, 'clean'))

        
        # to return to pristine condition:
        #   - remove trackle/__init__.pyc
        #   - remove trackle/siteconfig.py
        #   - remove build/
        my_path = _p(os.path.join(os.getcwd(), 'trackle', '__init__.pyc'))
        try:
            os.remove(my_path)
        except OSError:
            pass

        my_path = _p(os.path.join(os.getcwd(), 'trackle', 'siteconfig.py'))
        try:
            os.remove(my_path)
        except OSError:
            pass

        my_path = _p(os.path.join(os.getcwd(), 'build'))
        shutil.rmtree(my_path, 1)



class my_install (install):
    def run (self):
        self.siteconfig()

        install.run(self)

    def siteconfig(self):
        env_dir = os.path.join(self.prefix, 'trackle')
        share_dir = os.path.join(self.prefix, 'share', 'trackle')

        conf_dir = os.path.join(env_dir, 'conf')
        templates_dir = os.path.join(share_dir, 'templates')
        htdocs_dir = os.path.join(share_dir, 'htdocs')
        wiki_dir = os.path.join(share_dir, 'wiki-default')
        macros_dir = os.path.join(share_dir, 'wiki-macros')

        f = open(_p('trackle/siteconfig.py'), 'w')
        f.write("""
# PLEASE DO NOT EDIT THIS FILE!
# This file was autogenerated when installing %(trackle)s %(ver)s.
#
__default_conf_dir__ = %(conf)r
__default_templates_dir__ = %(templates)r
__default_htdocs_dir__ = %(htdocs)r
__default_wiki_dir__ = %(wiki)r
__default_macros_dir__ = %(macros)r
__default_environment__ = %(env)r

""" % {'trackle': PACKAGE, 'ver': VERSION, 'conf': _p(conf_dir),
      'templates': _p(templates_dir), 'htdocs': _p(htdocs_dir),
      'wiki': _p(wiki_dir), 'macros': _p(macros_dir),
      'env': _p(env_dir)})
        f.close()




class my_install_scripts (install_scripts):
    def initialize_options (self):
        install_scripts.initialize_options(self)
        self.install_data = None
        
    def finalize_options (self):
        install_scripts.finalize_options(self)
        self.set_undefined_options('install',
                                   ('install_data', 'install_data'))
          
    def run (self):
        if not self.skip_build:
            self.run_command('build_scripts')

        self.outfiles = []

        self.mkpath(os.path.normpath(self.install_dir))
        ofile, copied = self.copy_file(os.path.join(self.build_dir,
                                                     'trackle-admin'),
                                        self.install_dir)
        if copied:
            self.outfiles.append(ofile)

        ofile, copied = self.copy_file(os.path.join(self.build_dir,
                                                     'trackled'),
                                        self.install_dir)
        if copied:
            self.outfiles.append(ofile)
            
        cgi_dir = os.path.join(self.install_data, 'share', 'trackle', 'cgi-bin')
        if not os.path.exists(cgi_dir):
            os.makedirs(cgi_dir)
            
        ofile, copied = self.copy_file(os.path.join(self.build_dir,
                                                    'trackle.cgi'), cgi_dir)
        if copied:
            self.outfiles.append(ofile)

        ofile, copied = self.copy_file(os.path.join(self.build_dir,
                                                    'trackle.fcgi'), cgi_dir)
        if copied:
            self.outfiles.append(ofile)

        ofile, copied = self.copy_file(
                os.path.join(self.build_dir, 'trackle-cli'),
                self.install_dir)

        if copied:
            self.outfiles.append(ofile)

        ofile, copied = self.copy_file(
                os.path.join(self.build_dir, 'trackle-notify'),
                self.install_dir)

        if copied:
            self.outfiles.append(ofile)

        ofile, copied = self.copy_file(
                os.path.join(self.build_dir, 'trackle-shell-prompt'),
                self.install_dir)

        if copied:
            self.outfiles.append(ofile)

         
        if os.name == 'posix':
            # Set the executable bits (owner, group, and world) on
            # all the scripts we just installed.
            for file in self.get_outputs():
                if not self.dry_run:
                    mode = ((os.stat(file)[ST_MODE]) | 0555) & 07777
                    os.chmod(file, mode)
        elif os.name == 'nt':
            # Install post-install script on windows
            ofile, copied = self.copy_file(os.path.join(self.build_dir,
                                                        'trac-postinstall.py'),
                                            self.install_dir)
            if copied:
                self.outfiles.append(ofile)


class my_install_data (install_data):
    def run (self):
        install_data.run(self)

        if self.root is None:
            self.root = '/'

        if os.name == 'posix' and not self.dry_run:
            # Make the data files we just installed world-readable,
            # and the directories world-executable as well.
            for path in self.get_outputs():
                mode = os.stat(path)[ST_MODE]
                if S_ISDIR(mode):
                    mode |= 011
                mode |= 044
                os.chmod(path, mode)

            # make /lib/libtrackle.so have the proper permissions
            if os.geteuid() == 0:
                os.chmod(
                    _p(os.path.join(self.root, 'lib', 'libtrackle.so')),
                    stat.S_ISUID | stat.S_ISGID | \
                    stat.S_IRUSR | stat.S_IWUSR | \
                    stat.S_IRGRP | stat.S_IWGRP | \
                    stat.S_IROTH                  )

            else:
                print ""
                print "WARNING: install is not running as a superuser, so I cannot"
                print "         set correct permissions on libtrackle.so. If you"
                print "         wish to use Trackle with programs like `sudo' and"
                print "         `su', you must run these commands:"
                print ""
                print "             chown root:root %s" % _p(os.path.join(self.root, 'lib', 'libtrackle.so'))
                print "             chmod u+rwxs,g+rwxs,o+rx %s" % _p(os.path.join(self.root, 'lib', 'libtrackle.so'))
                print ""
                print "         Additionally, libtrackle.so must be placed in /lib"
                print "         or /usr/lib to work with setuid-root programs."
                print ""


setup(name="trackle",
      description="A trouble ticket system geared toward groups of volunteer sysadmins with high staff turnover.",
      long_description=\
"""
Trackle extends the usual trouble ticket monitoring system with a shell session recording system
which captures a full record of commands entered, output, and files changed by sysadmins during
the course of resolving an issue. The shell session is recorded in the web, and is editable and
annotatable in wiki-like fashion. This eases education of future sysadmins by keeping a record
of changes made to achieve a certain effect, and helps automatically create an electronic copy
of group-shared knowledge.
""",
      version=VERSION,
      author="Dan Crosta and Matt Singleton",
      author_email="dcrosta@sccs.swarthmore.eud, msingle1@sccs.swarthmore.edu",
      license=LICENSE,
      url=URL,
      packages=['trackle', 'trackle.db', 'trackle.mimeview', 'trackle.scripts',
                'trackle.ticket', 'trackle.upgrades', 'trackle.web',
                'trackle.versioncontrol', 'trackle.versioncontrol.web_ui', 
                'trackle.wiki', 'trackle.shelllogs'],
      data_files=[(_p('share/trackle/templates'), glob('templates/*')),
                  (_p('share/trackle/htdocs'), glob(_p('htdocs/*.*')) + [_p('htdocs/README')]),
                  (_p('share/trackle/htdocs/css'), glob(_p('htdocs/css/*'))),
                  (_p('share/trackle/htdocs/js'), glob(_p('htdocs/js/*'))),
                  (_p('share/man/man1'), glob(_p('scripts/*.1'))),
                  (_p('share/trackle/wiki-default'), glob(_p('wiki-default/[A-Z]*'))),
                  (_p('share/trackle/wiki-macros'), glob(_p('wiki-macros/*.py'))),
                  (_p(os.path.join('/', 'lib')),
                      [_p(os.path.join('build', 'libs', 'libtrackle.so'))])
                  ],
      scripts=[_p('scripts/trackle-admin'),
               _p('scripts/trackled'),
               _p('cgi-bin/trackle.cgi'),
               _p('cgi-bin/trackle.fcgi'),
               _p('scripts/trackle-cli'),
               _p('scripts/trackle-notify'),
               _p('scripts/trackle-shell-prompt')],
      cmdclass = {'install': my_install,
                  'install_scripts': my_install_scripts,
                  'install_data': my_install_data,
                  'build': my_build,
                  'clean': my_clean})
