import dtk
import dtk.util

from trackle.ticket.model import  Ticket
from trackle.core import  TracError
from trackle.sts import TrackleNoPermException, ShellTrackingSession

import time
import sys
import os, os.path
import random
from crypt import crypt


class TrackleCLI:
    """
    This file runs the main CLI entry point for trackle. Run as
    /path/to/trackle.py or python /path/to/trackle.py.
    """
    
    # these are the fields shown in the ticket editing/creating view
    #
    # the 'description' field is displayed separately, so should not be
    # included in this list
    fieldsToShow = [
        'summary',
        'type',
        'reporter',
        'owner',
        'priority',
        'severity',
        'component',
        'confirmed',
        'keywords',
        'cc',
        ]
    
    def __init__(self, env, username = ''):
        """
        Creates the TrackleCLI program (setting up dtk, etc) and runs the main loop.

        @author: Dan Crosta <dcrosta@sccs.swarthmore.edu>
        """

        self.env = env
        self.db = self.env.get_db_cnx()
    
        # username and password as authenticated from Trackle
        self.username = username
        self.password = ''
        validuser = False
    
        while not validuser:
            try:
                if self.username == '':
                    self.username = self.rawPromptForInput('Trackle username:')
                self.password = self.rawPromptForPassword('Trackle password:')

            except KeyboardInterrupt:
                print ''
                sys.exit()
        
            validuser = self.authenticate(self.username, self.password) 
    
            if not validuser:
                print "Invalid username or password"
                self.username = ''
                self.password = ''


    
    
        # now set up the engine and initial screen
        import logging
        self.engine = dtk.Engine('Trackle', log = False, loglevel = logging.INFO)
        self.engine.hideCursor()
    
    
        # create a stack for the main screen, so that we can push and pop screens
        self.main = dtk.Rows(self.engine, 'main', outerborder = False, innerborder = True)
        self.main.unbindKey('all')
        self.main.bindKey('q', self.engine.quit)

        # top row is instructions on how to use
        self.statusBar = InstructionBar(self.main, 'instructions', 'Loading...', [])
        self.showStatusBar('ticket list')

        self.main.addRow(self.statusBar, fixedsize = 1)

        self.stack = dtk.Stack(self.main, 'stack')
        self.main.addRow(self.stack, weight = 1)

        # the first screen is the ticket list, a TextTable showing summaries, owners, etc
        self.tlist = dtk.TextTable(self.stack, 'ticket list', vimlike = True, spacing = 1)

        # columns: ID, Summary, Owner, Changed (Date)
        self.tlist.addColumn(fixedsize = 4, alignment = 'right', name='ID')
        self.tlist.addColumn(weight = 2.0, name='Summary')
        self.tlist.addColumn(weight = 0.75, name = 'Owner')
        self.tlist.addColumn(fixedsize = 16, name = 'Changed')

        self.tlist.bindKey('enter', self.showTicketDetails)
        self.tlist.bindKey('focused', self.showStatusBar, 'ticket list')
        self.tlist.bindKey('n', self.createTicket)
        self.tlist.bindKey('r', self.refreshTicketList)

        # populate the ticket list
        self.refreshTicketList()

        self.stack.push(self.tlist)
    
        # set up dtk and run it
        self.engine.setRoot(self.main)
        self.engine.setFocus(self.tlist)
        self.engine.mainLoop()


    def authenticate(self, username, password):
        """
        Authenticates the user against the database of users stored by trackle.
        Currently this database is an apache htpasswd file (as created with
        the htpasswd tool), using salted/crypted passwords. The path to the
        file is given by the 'htpasswd' field in trackle.ini.
    
        @param username: the username of the user
        @type  username: string
    
        @param password: the password to go with the username
        @type  password: string
    
        @rtype: boolean
        @return: True if the username exists in the database and the password
            matches the username; False otherwise
        """
        htpasswd = self.env.config.get('trackle', 'userfile')
        if not os.path.isfile(htpasswd):
            raise TracError("Could not find user file '%s'" % htpasswd)

        fp = file(htpasswd, 'r')

        authenticated = False

        for line in fp.readlines():
            (un, pw) = line.split(':')
            pw = pw.strip()

            if username == un:
                salt = pw[:2]
                crypted = crypt(password, salt)

                if crypted == pw:
                    authenticated = True

                    break

        return authenticated


    def getTickets(self,status = ['new', 'assigned', 'reopened'], changedsince = None, changedbefore = None): 
        """
        Returns a list of Tickets that exist in the trac database with
        any of the given statuses. By default, lists "open" tickets.

        @param status: a list of status types to return
        @type  status: list of strings

        @param changedsince: an integer time (eg, time.time()) which
            limits results only to tickets changed after this time
        @type  changedsince: integer

        @param changedbefore: an integer time (eg, time.time()) which
            limits results only to tickets changed before this time
        @type  changedbefore: integer

        @rtype: list of Tickets
        @return: a list of Tickets that exist in the trac database
        """
        # the db cursor
        cursor = self.db.cursor() 

        args = []
        sql =  "SELECT id FROM ticket "
        where = []

        if len(status):
            where.append("status IN (%s)" % ', '.join(["'%s'" for elm in status]))
            args.extend(status)
        if changedsince:
            where.append("changetime > %d")
            args.append(int(changedsince))
        if changedbefore:
            where.append("changetime < %d")
            args.append(int(changedbefore))

        if len(where):
            sql += "WHERE " + ' AND '.join(where)

        args = tuple(args)

        cursor.execute(sql % args)
        rows = cursor.fetchall()

        # create the list of Tickets to return
        out = []
        for row in rows:
            out.append(Ticket(self.env, tkt_id = row[0]))

        return out


    def refreshTicketList(self):
        tickets = self.getTickets()

        # info and ids are used to populate the list of tickets
        info = []
        ids = []

        for ticket in tickets:
            info.append(
                    [ticket.id,
                    ticket['summary'],
                    ticket['owner'],
                    time.strftime('%d %b %y %H:%M', time.localtime(int(ticket.time_changed)))
                    ])
            ids.append(ticket.id)

        self.tlist.setItems(info, ids)


    def showStatusBar(self, whatfor):
        if whatfor == 'ticket list':
            self.statusBar.setTitle('Open Tickets >>')
            self.statusBar.setInstructions(
                    [('up, dn', 'navigate'),
                     ('enter', 'select ticket'),
                     ('n', 'new ticket'),
                     ('r', 'refresh list'),
                     ('q', 'quit')
                     ])

        elif whatfor == 'ticket view':
            self.statusBar.setTitle('Viewing >>')
            self.statusBar.setInstructions(
                    [('b', 'back to ticket list'),
                     ('s', 'start shell session'),
                     ('u', 'unlock ticket'),
                     ('e', 'edit ticket'),
                     ('q', 'quit')
                     ])

        elif whatfor == 'ticket editing':
            self.statusBar.setTitle('Editing >>')
            self.statusBar.setInstructions(
                    [('up, dn', 'navigate fields'),
                     ('enter', 'edit field'),
                     ('d', 'switch to description'),
                     ('s', 'save'),
                     ('esc', 'cancel'),
                     ])

        elif whatfor == 'ticket creation':
            self.statusBar.setTitle('New Ticket >>')
            self.statusBar.setInstructions(
                    [('up, dn', 'navigate fields'),
                     ('enter', 'edit field'),
                     ('d', 'switch to description'),
                     ('s', 'save'),
                     ('esc', 'cancel'),
                     ])

        elif whatfor == 'text field editing':
            self.statusBar.setInstructions(
                    [('esc/enter', 'back to field selection')])

        elif whatfor == 'select field editing':
            self.statusBar.setInstructions([
                    ('esc/enter', 'back to field selection'),
                    ('right', 'next option'),
                    ('left', 'previous option'),
                    ])

        elif whatfor == 'finish shell session':
            self.statusBar.setTitle('Files to Include >>')
            self.statusBar.setInstructions([
                    ('space', 'toggle select'),
                    ('up, dn', 'navigate'),
                    ('s', 'save and exit'),
                    ('c', ('exit without saving', 'cancel'))
                    ])


    def createTicket(self):
        """
        create a new Ticket object and hand it off to _showTicketDetails
        and go into edit mode
        """
        # this creates a new Ticket with default values
        ticket = Ticket(self.env)

        self.env.log.debug('Created new ticket %s with ID: %s' % (ticket, ticket.id))

        ticket['status'] = 'new'
        ticket['reporter'] = self.username
        ticket['description'] = "Hit 'd' to change the default description"

        self._showTicketDetails(ticket)
        self.beginEditTicket(ticket = ticket)


    def showTicketDetails(self):
        """
        get the ticket ID from the ticket list, create a Ticket,
        then call _showTicketDetails() with that ticket object.
        """
        # get the selected ticket ID
        id = self.tlist.index()


        # update & get the ticket
        ticket = Ticket(self.env, id)

        return self._showTicketDetails(ticket)

    
    def _showTicketDetails(self, ticket):
        """
        create dtk widgets to display the details of a ticket and set up
        keybindings
        """

        self.details = dtk.Rows(self.stack, 'ticket details', outerborder = False, innerborder = False)

        width = max([len(field) for field in self.fieldsToShow])
        width = max([len(field['label']) for field in ticket.fields if field['name'] in self.fieldsToShow] + [width])

        fmt = '%%%ds ' % width
        width += 1 # for the trailing space

        # dict of field name => editor for later use
        # saving tickets, etc
        self.fieldEditors = {}
        
        self.ticketFields = dtk.Rows(self.details, 'ticketFields', innerborder = False, outerborder = False)
        self.ticketFields.unbindKey('tab')

        # create a dictionary of name => field (dict)
        fields = {}
        for field in ticket.fields:
            name = field['name']
            fields[name] = field

        for name in self.fieldsToShow:

            # this happens when a select field has no
            # values set, so it just won't show up
            if name not in fields:
                continue

            field = fields[name]

            if name not in self.fieldsToShow:
                continue

            if 'label' in field:
                title = field['label']
            else:
                title = field['name'].title()

            row = dtk.Columns(self.ticketFields, 'details:row:%s' % name, outerborder = False, innerborder = False)
            row.unbindKey('tab')

            label = dtk.Button(row, 'label:%s' % name, title, decoration = fmt)
            row.addColumn(label, fixedsize = width)
            row.addSeparator(type = 'line')
            row.addSeparator(type = 'blank')

            if field['type'] in ('text',):
                 editor = dtk.TextField(row, 'field:%s' % name)
                 if name in ticket.values:
                     editor.setText(ticket[name])

            elif field['type'] in ('radio', 'select'):
                editor = dtk.Select(row, 'field:%s' % name)
                options = field['options']
                if 'optional' in field and field['optional']:
                    options.insert(0, '')
                options.reverse()
                editor.setOptions(options)
                if name in ticket.values:
                    editor.setValue(ticket[name])

            row.addColumn(editor)

            row.bindKey('focused', self.engine.setFocus, label)
            row.bindKey('down', self.ticketFields.nextRow)
            row.bindKey('j', self.ticketFields.nextRow)
            row.bindKey('up', self.ticketFields.prevRow)
            row.bindKey('k', self.ticketFields.prevRow)

            label.bindKey('click', self.beginEditField, row, editor, field)
            editor.bindKey('esc',  self.stopEditField,  row)
            editor.bindKey('enter',  self.stopEditField,  row)

            self.fieldEditors[name] = editor

            self.ticketFields.addRow(row, fixedsize = 1)

        self.details.addRow(self.ticketFields, fixedsize = len(self.fieldEditors))

        self.details.addSeparator()

        description = ''
        try:
            description = ticket['description'].replace('\r', '')
        except:
            pass

        self.desc = dtk.Pager(self.details, 'description', vimlike = True)
        self.desc.setText(description)
        self.details.addRow(self.desc, weight = 1)

        id = ticket.id

        self.details.bindKey('s', self.startShellSession, id)
        self.details.bindKey('b', self.stack.pop, forceFocus = True)
        self.details.bindKey('u', self.unlockTicket, id)
        self.details.bindKey('e', self.beginEditTicket, id, ticket.time_changed)
        self.details.unbindKey('tab')
        
        # now show it
        self.stack.push(self.details)

        # set the focus on the description pane so scrolling works
        self.engine.setFocus(self.desc)

        # update the status bar
        self.showStatusBar('ticket view')


    def beginEditTicket(self, id = None, time_changed = None, ticket = None):
        if ticket is not None:
            self.showStatusBar('ticket creation')
            self.ticketFields.bindKey('esc', self.stopEditTicket, mode = 'return')
            self.ticketFields.bindKey('s', self.stopEditTicket, ticket = ticket)

        else:
            self.showStatusBar('ticket editing')
            self.ticketFields.bindKey('esc', self.stopEditTicket, mode = 'cancel')
            self.ticketFields.bindKey('s', self.stopEditTicket, id = id, time_changed = time_changed)

        self.ticketFields.switchRow(0)
        self.details.bindKey('d', self.startTextEditor, self.desc)


    def stopEditTicket(self, id = None, time_changed = None, ticket = None, checkpoint = 'start', mode = 'save'):
        if mode in ('cancel', 'return'):
            # return to ticket view
            self.showStatusBar('ticket view')
            self.ticketFields.unbindKey('esc')
            self.ticketFields.unbindKey('s')
            self.details.unbindKey('d')
    
            self.engine.setFocus(self.desc)

            if mode is 'return':
                self.stack.pop(forceFocus = True)

            return

        if id is not None:
            # id will be None for new tickets. if it is not set, then
            # overwrite ticket (if it's set) with the updated ticket
            ticket = Ticket(self.env, id)
    
            if ticket.time_changed != time_changed:
                # XXX: hack!
                self.stack.pop()
                self._showTicketDetails(ticket)
    
                self.prompt('Collision',
                    'Someone else has changed the ticket while you were editing! '
                    'Your changes have been discarded. ')
                return

        # description is handled separately
        fields = [field['name'] for field in ticket.fields if field['name'] in self.fieldsToShow and field['name'] != 'description']

        for field in fields:
            editor = self.fieldEditors[field]
            value = editor.getText()

            ticket[field] = value

        ticket['description'] = self.desc.getText()

        if ticket.is_changed():
            if id is not None:
                # then this is an existing ticket and we need
                # a change message

                # prepare the change message prompt
                class FauxField:
                    def __init__(self):
                        self.text = '\n'.join([
                                '',
                                '---- Everything below and including this line will be ignored ----',
                                '',
                                'You must enter a ticket change message to be appended to the',
                                'ticket change history in Trackle. You may use WikiFormatting.',
                                ])
                    def setText(self, text):
                        self.text = text
                    def getText(self):
                        return self.text
                    def getResult(self):
                        lines = self.text.split('\n')
        
                        for i, line in enumerate(lines):
                            if line.strip() == '---- Everything below and including this line will be ignored ----':
                                break
                        
                        return '\n'.join(lines[:i])
        
                # prompt for the change message
                prompt = FauxField()
                self.startTextEditor(prompt)
        
                msg = prompt.getResult()
        
                ticket.save_changes(author = self.username, comment = msg)

            else:
                # this is a new ticket and we just need to
                # insert it into the database
                id = ticket.insert()
                self.refreshTicketList()

                # refresh the ticket detail with the new ticket (with ID)
                self.stack.pop()
                self._showTicketDetails(ticket)


        # return to ticket view
        self.showStatusBar('ticket view')
        self.ticketFields.unbindKey('esc')
        self.details.unbindKey('d')

        self.engine.setFocus(self.desc)

        
    def beginEditField(self, row, editor, field):
        row.nextColumn()

        if field['type'] in ('text',):
            self.showStatusBar('text field editing')
            editor.moveToEnd()

        elif field['type'] in ('select', 'radio'):
            self.showStatusBar('select field editing')



    def stopEditField(self, row):
        self.showStatusBar('ticket editing')
        self.engine.hideCursor()
        row.prevColumn()


    def unlockTicket(self, id, checkpoint = 'start'):
        """
        This is to unlock a ticket without starting a shell session.
        """

        # update the ticket
        ticket = Ticket(self.env, id)

        if ticket.is_locked():
            if checkpoint is not 'unlock':
                self.promptForInput(
                        'Ticket Locked',
                        'This ticket is locked by %s, with code %s. THIS PROBABLY MEANS '
                        'SOMEONE IS ACTIVELY WORKING ON THIS TICKET! If you are SURE '
                        'you want to unlock this ticket, type the four digit lock '
                        'code below and hit enter.' % (ticket['owner'], ticket['locked']),
                        callback = self.unlockTicket, id = id, checkpoint = 'unlock')
                return

            else: # if checkpoint == 'unlock'
                codeEntered = self.dlg.getResult()
                if codeEntered != ticket['locked']:
                    self.prompt('Wrong Lock Code', 'The lock code you entered does not '
                            'match the lock code on this ticket. Returning to ticket '
                            'detail.')

                else:
                    ticket.unlock(codeEntered, self.username, 'Unlocking ticket from trackle-cli.')
                    self._showTicketDetails(ticket)


    def startTextEditor(self, field):
        """
        field must support setText() and getText()
        """

        # see if the user has VISUAL or EDITOR set, if not, try to find vim, emacs or nano
        editor = None
        if editor is None and os.environ.has_key('VISUAL'):
            editor = os.environ['VISUAL']

        elif editor is None and os.environ.has_key('EDITOR'):
            editor = os.environ['EDITOR']

        else:
            # try the usual places
            paths = [
                os.path.join('/','usr','bin','vim'),
                os.path.join('/','usr','local','bin','vim'),
                os.path.join('/','usr','bin','emacs'),
                os.path.join('/','usr','local','bin','emacs'),
                os.path.join('/','usr','bin','nano'),
                os.path.join('/','usr','local','bin','nano'),
                ]
            for path in paths:
                if os.path.exists(path) and os.path.isfile(path) and os.access(path, os.R_OK | os.X_OK): 
                    editor = path
                    break

        if editor is None:
            self.prompt('No Editor',
                    'You have no VISUAL or EDITOR environment variable set, '
                    'nor could Trackle find vim, emacs or nano in any of the '
                    'usual places. Quit Trackle and set VISUAL or EDITOR to '
                    'the path to a text editor of your choice.')
            return


        # change the window title to indicate we're in an STS
        oldtitle = self.engine.getTitle()
        self.engine.setTitle('Trackle: Edit Description Field')

        # drop to shell mode and execute the STS
        self.engine.shellMode()

        # create the temporary file
        filename = self.tempFilename()

        fp = file(filename, 'w')
        fp.writelines(field.getText())
        fp.close()


        # start the editor
        os.spawnl(os.P_WAIT, editor, editor, filename)


        # get the results
        fp = file(filename, 'r')
        field.setText(''.join(fp.readlines()))
        fp.close()

        # clean up
        os.remove(filename)

        # once that returns, return to dtk mode
        self.engine.dtkMode()

        # reset old title
        self.engine.setTitle(oldtitle)
        

    def startShellSession(self, id, checkpoint = 'start'):
        """
        Begins a shell session on the given ticket (by id)

        @param id: the numeric ID of the ticket to run the shell session on
        @type  id: int

        @param checkpoint: the stage in the multi-callback-nonsense we're at.
            when calling startShellSession directly, leave this as 0
        @type  checkpoint: steal
        """

        # we have to use some ugly callback chaining hacks to get the "am I able to start
        # a shell session?" logic to work, until dtk supports dialogs that block (...someday.)
        #
        # so:
        #
        # before we can start the shell session, update the ticket fields,
        # then we have a few tests to pass before beginning:
        #
        # if the ticket is locked (regardless of who locked it), we
        #   return without beginning the shell session
        #
        # if the ticket is closed (how did we get here?) we say so and
        #   return
        #
        # if the ticket is assigned and accepted by someone else,
        #   we should prompt for 'stealing' the ticket
        #
        # if we reach this point, we assign and accept the ticket
        #   (if necessary), and lock it and begin the shell session

        # checkpoints:
        #
        #  - start
        #  - steal (if the ticket should be stolen)
        #  - unlock (if the ticket should be unlocked)
        #  - shell (to start the shell session)

        # this will update the ticket
        ticket = Ticket(self.env, id)

        if ticket.is_locked():
            if checkpoint is not 'unlock':
                self.promptForInput(
                        'Ticket Locked',
                        'This ticket is locked by %s, with code %s. THIS PROBABLY MEANS '
                        'SOMEONE IS ACTIVELY WORKING ON THIS TICKET! If you are SURE '
                        'you want to begin a shell session, type the four digit lock '
                        'code below and hit enter.' % (ticket['owner'], ticket['locked']),
                        callback = self.startShellSession, id = id, checkpoint = 'unlock')
                return

            else: # if checkpoint == 'unlock'
                codeEntered = self.dlg.getResult()
                if codeEntered != ticket['locked']:
                    self.prompt('Wrong Lock Code', 'The lock code you entered does not '
                            'match the lock code on this ticket. Returning to ticket '
                            'detail.')
                    return

                else:
                    ticket.unlock(codeEntered, self.username, "Unlocking ticket to begin shell tracking session.")
                    time.sleep(1)
                    checkpoint = 'shell'
            

        if checkpoint == 'start':
            if ticket['status'] == 'assigned':
                if ticket['owner'] != self.username:
                    self.promptYesNo('Already Assigned',
                            'This ticket is assigned to %s, but not locked. '
                            'Are you sure you want to begin a shell session?' % ticket['owner'], 
                            callback = self.startShellSession, id = id, checkpoint = 'steal')

                    return

            elif ticket['status'] == 'new':
                pass
    
            elif ticket['status'] == 'reopened':
                pass
    
            elif ticket['status'] == 'closed':
                self.prompt('This ticket is closed. Re-open it before beginning a shell session.')
                return

            checkpoint = 'shell'


    
        if checkpoint == 'steal':
            steal = self.dlg.getResult()

            if not steal:
                self.prompt('OK', 'OK, returning to ticket detail.')
                return

            checkpoint = 'shell'


        # final check before beginning
        if checkpoint != 'shell':
            return


        # take over the ticket if we've gotten this far
        ticket['owner'] = self.username
        ticket['status'] = 'assigned'


        # this will set the above two fields as well
        lockKey = ticket.lock(self.username, 'Locking ticket for Shell Tracking Session')
        if lockKey is False:
            self.engine.dtkMode()
            return
        else:
            self.lockKey = lockKey

        # change the window title to indicate we're in an STS
        oldtitle = self.engine.getTitle()
        self.engine.setTitle('Trackle Shell Tracking Session')

        # drop to shell mode and execute the STS
        self.engine.shellMode()

        startTime = time.time()

        try:
            sts = ShellTrackingSession(int(id), self.env, ticket, self.username)
            (sessionID, fileList, history) = sts.run()

        except TrackleNoPermException, dir:
            print "Could not write %s" % dir
            sessionID = False

        endTime = time.time()
        if endTime - startTime < 1:
            # avoid an ugly error from the database if
            # the session lasted less than 1 second
            #
            # XXX: eventually will want to do something
            # smart here like show an error to the user,
            # probably.
            time.sleep(1)

        # once that returns, return to dtk mode
        self.engine.dtkMode()

        # reset old title
        self.engine.setTitle(oldtitle)

        if len(fileList):
            # show the file selection list
            files = dtk.ListBox(self.stack, 'files', vimlike=True)
            files.setDrawStyle(scheck = '+ ')
            fileList.sort()
            files.setItems(fileList)

            files.bindKey('s', self.finishShellSession, ticket, sts, lockKey, files)
            files.bindKey('c', self.cancelShellSession, ticket, sts, lockKey)
            files.bindKey('q', self.engine.ignore)

            self.showStatusBar('finish shell session')

            self.stack.push(files)
            self.engine.setFocus(files)

        # otherwise, only keep the session if the history
        # contains more than an immediate 'exit' command
        elif len(history) > 1:
            sessionID = sts.makeWikiPage(fileList)
            ticket.unlock(lockKey, self.username, 'Unlocking ticket after Shell Tracking Session\n\nNew [shellsession:%s Shell Session (%s)] recorded' % (sessionID, sessionID))

        else:
            ticket.unlock(lockKey, self.username, 'Shell session aborted.')


    def cancelShellSession(self, ticket, sts, lockKey):
        ticket.unlock(lockKey, self.username, 'Unlocking ticket after cancelled Shell Tracking Session')

        self.stack.pop()
        self.stack.pop()
        self._showTicketDetails(ticket)
        
        sts.cleanup()


    def finishShellSession(self, ticket, sts, lockKey, files):
        fileList = files.getSelectedItems()

        sessionID = sts.makeWikiPage(fileList)
        ticket.unlock(lockKey, self.username, 'Unlocking ticket after Shell Tracking Session\n\nNew [shellsession:%s Shell Session (%s)] recorded' % (sessionID, sessionID))

        self.stack.pop()
        self.stack.pop()
        self._showTicketDetails(ticket)

        # clean up session dir
        sts.cleanup()


    
    def rawPromptForPassword(self, msg = 'Password:'):
        """
        mostly copied from <http://www.python.org/doc/2.4.2/lib/node433.html>
        """
        import termios, sys
    
        fd = sys.stdin.fileno()
    
        old = termios.tcgetattr(fd)
        new = termios.tcgetattr(fd)
    
        # turn off terminal echoing
        new[3] = new[3] & ~termios.ECHO
    
        try:
            termios.tcsetattr(fd, termios.TCSADRAIN, new)
            password = raw_input(msg + ' ')
            print ""
    
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)
    
        return password
    
    
    def rawPromptForInput(self, msg = 'Username:'):
        """
        simple method to get a string from the terminal
        """
        username = raw_input(msg + ' ')
    
        return username


    def _getDialog(self):
        self.dlg = getattr(self, 'dlg', None)

        if self.dlg is None:
            self.dlg = dtk.Dialog(self.engine, 'dlg')

        return self.dlg
        
    def prompt(self, title, msg, *args, **kwargs):
        dlg = self._getDialog()

        dlg.setType('message')
        dlg.setText(msg)
        dlg.setTitle(title)

        dlg.unbindKey('all')
        if 'callback' in kwargs:
            callback = kwargs['callback']
            del kwargs['callback']

            dlg.bindKey('dismissed', callback, *args, **kwargs)


        dlg.show()


    def promptYesNo(self, title, msg, *args, **kwargs):
        dlg = self._getDialog()

        dlg.setTitle(title)
        dlg.setText(msg)
        dlg.setType('yesno')

        dlg.unbindKey('all')
        if 'callback' in kwargs:
            callback = kwargs['callback']
            del kwargs['callback']

            dlg.bindKey('dismissed', callback, *args, **kwargs)

        dlg.show()

    
    def promptForInput(self, title, msg, *args, **kwargs):
        dlg = self._getDialog()

        dlg.setTitle(title)
        dlg.setText(msg)
        dlg.setType('input')

        dlg.unbindKey('all')
        if 'callback' in kwargs:
            callback = kwargs['callback']
            del kwargs['callback']

            dlg.bindKey('dismissed', callback, *args, **kwargs)

        dlg.show()


    def tempFilename(self):
        """
        return a filename suitable for writing temporary things
        to, and make the directory it should live in, if that
        doesn't exist
        """
        filename = os.path.join(self.env.config.get('trackle', 'tempdir'), 'sessions-%s' % os.environ['USER'], 'trackle-%0d.tmp' % random.randint(0000, 9999))

        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

        return filename
         

class InstructionBar(dtk.Drawable):
    
    def __init__(self, parent, name, title, instructions):
        super(InstructionBar, self).__init__(parent, name)

        self.setTitle(title)
        self.setInstructions(instructions)


    def setTitle(self, title):
        self.title = title
        self.touch()


    def setInstructions(self, instructions):
        self.instructions = instructions
        self.touch()
    

    def _getExplanation(self, explanation, short = False):
        """
        returns the explanation or a shortened version
        of it. when operating in short mode,
        if the explanation is a string, the short
        explanation is the first word of the string.
        if it is a tuple, take the shortest element

        @param short: whether to use short mode
        @type  short: boolean
        """
        import types

        if short:
            if type(explanation) in types.StringTypes:
                return explanation.split()[0]

            else:
                shortest = None

                for exp in explanation:
                    if shortest is None or len(exp) < len(shortest):
                        shortest = exp

                return shortest

        else:
            if type(explanation) in types.StringTypes:
                return explanation

            else:
                return explanation[0]


    def render(self):
        length  = sum([len(key) + len(self._getExplanation(exp)) for (key, exp) in self.instructions])
        length += len(self.instructions) * 4
        length += len(self.title) + 2

        short = False
        if length > self.w:
            # now we try the shortened version
            length  = sum([len(key) + len(self._getExplanation(exp, True)) for (key, exp) in self.instructions])
            length += len(self.instructions) * 4
            length += len(self.title) + 2

            if length > self.w:
                raise Exception, "more space is required than available" 

            short = True

        self.draw(' ' + self.title + ' ', 0, 0, foreground='black', background='white')

        col = len(self.title) + 2
        for (key, explanation) in self.instructions:
            explanation = self._getExplanation(explanation, short)

            self.draw(key, 0, col, bold = True, foreground='black', background='white')
            col += len(key)
            self.draw(': %s  ' % explanation, 0, col, foreground='black', background='white')
            col += len(explanation) + 4

        self.draw(' ' * (self.w - col), 0, col, foreground='black', background='white')
