import trackle.env
from trackle.ticket.notification import  PeriodicNotifyEmail
from trackle.ticket.model import Ticket

import time
import os.path

def getRunfileName(env):
    """
    return the name of the last-run-time file
    """
    return os.path.join(env.config.get('trackle', 'statedir'), 'notify-last-ran')


def getLastRunTime(env):
    """
    get the last-run time from the state file
    """
    runfile = getRunfileName(env)

    try:
        fp = file(runfile, 'r')
        lasttime = fp.readline().strip()
        fp.close()
        lasttime = int(lasttime)

    except:
        lasttime = 0

    return lasttime


def updateLastRunTime(env):
    """
    update the last-run time in the state file
    """
    runfile = getRunfileName(env)

    fp = file(runfile, 'w')
    fp.write('%d\n' % int(time.time()))
    fp.close()


def getEnvironment():
    import trackle.siteconfig

    env = trackle.env.Environment(trackle.siteconfig.__default_environment__)

    return env

def run():
    env = getEnvironment()

    runtime = getLastRunTime(env)

    db = env.get_db_cnx()
    cursor = db.cursor()

    # first get all open tickets
    sql = """SELECT id
      FROM ticket
      WHERE status <> 'closed'
      ORDER BY id ASC"""
    rows = cursor.execute(sql)
    rows = cursor.fetchall()

    openTickets = []
    for row in rows:
        openTickets.append(Ticket(env, tkt_id = row[0]))


    # next get all tickets changes since lastruntime
    
    # this complicated SQL does one cool thing: it joins
    # ticket_change with itself so that it can get, in the
    # same row, the action (change to ticket status) and
    # also the resolution message, *if* the action was
    # to close the ticket, otherwise the 'resolution'
    # column is blank for the row
    sql = """SELECT DISTINCT
        a.ticket,
        a.time,
        a.author,
        a.newvalue AS action,
        (CASE a.newvalue
          WHEN 'closed' THEN
            (CASE b.field WHEN 'resolution' THEN b.newvalue END)
          WHEN 'assigned' THEN
            (CASE b.field WHEN 'owner' THEN b.newvalue END)
        END) AS resolution 
      FROM ticket_change a, ticket_change b
      WHERE
        a.time >= %d AND
        a.ticket = b.ticket AND
        a.time = b.time AND
        a.field='status' AND
        ((a.newvalue='closed' AND b.field='resolution') OR
         (a.newvalue='assigned' AND b.field='owner') OR
         a.newvalue IN ('new', 'reopened'))
      ORDER BY a.ticket, a.time"""

    cursor.execute(sql % runtime)
    rows = cursor.fetchall()

    lastid = None
    changelog = None
    ticketChanges = []

    for row in rows:
        (id, time, author, action, value) = row

        if id is not lastid:
            if changelog is not None:
                ticketChanges.append(changelog)

            changelog = TicketChangelog()
            changelog.ticket = Ticket(env, tkt_id = id)

        # print '* adding changeset (%d, %s, %s, %s) to %d' % (time, author, action, value, changelog.ticket.id)
        changelog.append(ChangeSet(time, author, action, value))

        lastid = id


    # now we have a list of ticket status changes grouped by ticket,
    # which is all ready to be handed off to PeriodicNotifyEmail
    # and sent off to staff
    eml = PeriodicNotifyEmail(env) 
    eml.notify(openTickets, ticketChanges, runtime)


    updateLastRunTime(env)


class TicketChangelog:
    def __init__(self):
        self.ticket = None
        self.changes = []

    def append(self, thing):
        self.changes.append(thing)

class ChangeSet:
    def __init__(self, time, author, action, value):
        self.time = time
        self.author = author
        self.action = action
        self.value = value
