# -*- coding: iso-8859-1 -*-
#
# Copyright (C) 2003-2005 Edgewall Software
# Copyright (C) 2003-2005 Jonas Borgstrm <jonas@edgewall.com>
# Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Jonas Borgstrm <jonas@edgewall.com>
#         Christopher Lenz <cmlenz@gmx.de>

from trackle.core import  *
from trackle.shelllogs.api import statToDict, interpretStats

class ShellLog(object):
    """Represents a shell log (new or existing)."""

    def __init__(self, env, ticket, seq=None, version=None, db=None):
        self.env = env
        self.ticket = int(ticket)
        self.text = None

        # if seq is None, we're creating the ShellLog for the first time
        if seq is None:
            self.seq = self._get_next_seq(db)

            self.text = None
            self.version = 0
            self.start = -1
            self.finish = -1

        else:
            self.seq = int(seq)
            self._fetch(ticket, seq, version, db)

        self.old_text = self.text

    def _fetch(self, ticket, seq, version, db):
        if not db:
            db = self.env.get_db_cnx()
        cursor = db.cursor()
        if version:
            cursor.execute("SELECT text,version,start,finish FROM shellsession "
                           "WHERE ticket=%d AND seq=%d AND version=%d",
                           (int(ticket), int(seq), int(version)))
        else:
            # fetch the latest
            cursor.execute("SELECT text,version,start,finish FROM shellsession "
                           "WHERE ticket=%d AND seq=%d ORDER BY version DESC LIMIT 1",
                           (int(ticket), int(seq)))
        row = cursor.fetchone()
        
        if row is not None:
            text,version,start,finish = row
            self.text = text
            self.version = int(version)
            self.start = start
            self.finish = finish
        else:
            self.text = None
            self.version = 0
            self.start = -1
            self.finish = -1

    def _get_next_seq(self, db):
        if not db:
            db = self.env.get_db_cnx()
        cursor = db.cursor()

        cursor.execute("SELECT MAX(seq) FROM shellsession WHERE ticket=%d", (self.ticket, ))
        row = cursor.fetchone()

        if row[0] is not None:
            return int(row[0]) + 1
        else:
            return 1

    def save(self, author, remote_addr, db=None):
        if not db:
            db = self.env.get_db_cnx()
            handle_ta = True
        else:
            handle_ta = False

        self.version += 1
        
        if self.text != self.old_text:
            cursor = db.cursor()
            cursor.execute("INSERT INTO shellsession (ticket,seq,version,author,ipnr,"
                           "start,finish,text) VALUES (%d,%d,%d,%s,%s,%s,%s,%s)",
                           (self.ticket, self.seq, self.version, author,
                           remote_addr, self.start, self.finish, self.text))
        else:
            raise TracError('Page not modified')

        if handle_ta:
            db.commit()

        self.old_text = self.text



class FileDiff(object):
    """Represents a session file diff (new or existing)"""

    def __init__(self, env, ticket, seq, path, new = False, db=None):
        self.env = env

        self.ticket = ticket
        self.seq = seq
        self.path = path

        if not new:
            self._fetch(db)
                
    def _fetch(self, db):
        
        if not db:
            db = self.env.get_db_cnx()
        cursor = db.cursor()
        cursor.execute("SELECT start, finish, active, sstat, fstat "
                       "FROM filediff WHERE ticket=%d AND seq=%d AND path=%s",
                       (self.ticket, self.seq, self.path))
        row = cursor.fetchone()
        
        if row:
            start, finish, active, sstat, fstat = row

            self.start = start
            self.finish = finish
            self.active = active
            self.sstat = sstat
            self.fstat = fstat
            
        else:
            self.start = None
            self.finish = None
            self.active = 0
            self.sstat = None
            self.fstat = None
            

    def save(self, db=None):
        if not db:
            db = self.env.get_db_cnx()
            handle_ta = True
        else:
            handle_ta = False

        cursor = db.cursor()

        self.env.log.debug("INSERT INTO filediff (ticket,seq,path,start,finish,sstat,fstat,active)"
                           "VALUES (%d, %d, '%s', '%s', '%s', '%s', '%s', %d)" %
                           (self.ticket, self.seq, self.path, self.start, self.finish, self.sstat, self.fstat, self.active))

        cursor.execute("INSERT INTO filediff (ticket,seq,path,start,finish,sstat,fstat,active)"
                       "VALUES (%d,%d,%s,%s,%s,%s,%s,%d)",
                       (self.ticket, self.seq, self.path, self.start, self.finish,
                        self.sstat, self.fstat, self.active))

        if handle_ta:
            db.commit()
    
    def updateActive(self, remote_addr, db=None):
        if not db:
            db = self.env.get_db_cnx()
            handle_ta = True
        else:
            handle_ta = False

        cursor = db.cursor()

        cursor.execute("UPDATE filediff SET active=%d "
                       "WHERE ticket=%d AND seq=%d AND path=%s",
                       (int(self.active), self.ticket, self.seq, self.path))

        if handle_ta:
            db.commit()


    def interpret(self):
        """
        interpret what happened to the file by parsing the saved
        .stat and .stat.end contents (sstat and fstat)

        see shelllogs.api.interpretStats() for possible return values
        """

        # translate them to dicts
        sstat = statToDict(self.sstat)
        fstat = statToDict(self.fstat)


        return interpretStats(sstat, fstat)
