# -*- coding: iso-8859-1 -*-
#
# Copyright (C) 2003-2006 Edgewall Software
# Copyright (C) 2003-2005 Jonas Borgstrm <jonas@edgewall.com>
# Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Jonas Borgstrm <jonas@edgewall.com>
#         Christopher Lenz <cmlenz@gmx.de>

import re
import StringIO
from time import ctime
import md5

from trackle.core import  *
from trackle.util import  Markup, get_reporter_id, shorten_line
from trackle.perm import  IPermissionRequestor
from trackle.versioncontrol.diff import  get_diff_options, hdf_diff
from trackle.web.chrome import  add_link, add_stylesheet, INavigationContributor
from trackle.web import  IRequestHandler
from trackle.wiki import  IWikiSyntaxProvider
from trackle.wiki.formatter import  wiki_to_html, wiki_to_oneliner
from trackle.shelllogs.model import  ShellLog, FileDiff
from trackle.Search import  ISearchSource, search_to_sql, shorten_result

from trackle.ticket.model import  Ticket

import trackle.shelllogs.api

class ShellLogsModule(Component):

    implements(INavigationContributor, IPermissionRequestor, IRequestHandler,
            ISearchSource)

    # ISearchSource methods

    def get_search_filters(self, req):
        """
        Return a list of filters that this search source supports. Each
        filter must be a (name, label) tuple, where `name` is the internal
        name, and `label` is a human-readable name for display.
        """
        if req.perm.has_permission('SHELLLOGS_VIEW'):
            yield ('shellsession', 'Shell Sessions')
        

    def get_search_results(self, req, terms, filters):
        """
        Return a list of search results matching `terms`. The `filters`
        parameters is a list of the enabled
        filters, each item being the name of the tuples returned by
        `get_search_events`.

        The events returned by this function must be tuples of the form
        (href, title, date, author, excerpt).
        """
        if not 'shellsession' in filters:
            return

        db = self.env.get_db_cnx()
        cursor = db.cursor()

        sql_query, args = search_to_sql(db,
                ['s1.author', "s1.ticket || '-' || s1.seq", 's1.text'],
                terms)

        cursor.execute("SELECT s1.author, s1.ticket || '-' || s1.seq, s1.text, s1.finish "
                       "FROM shellsession s1, "
                       "(SELECT ticket, seq, max(version) AS ver "
                       " FROM shellsession s2 GROUP BY ticket, seq) s2 "
                       "WHERE s1.version = s2.ver AND s1.ticket = s2.ticket AND s1.seq = s2.seq AND "
                       + sql_query, args)

        # store partial results
        results = [(author, sid, text, time) for (author, sid, text, time) in cursor]
        sids = [sid for (author, sid, text, time) in results]


        sql_query, args = search_to_sql(db, ['f.path', 'f.start', 'f.finish'], terms)

        cursor.execute("SELECT s.author, s.ticket || '-' || s.seq, f.path, f.start, f.finish, s.finish "
                       "FROM shellsession s, filediff f "
                       "WHERE f.ticket = s.ticket AND f.seq = s.seq AND f.active = 1 AND " +
                       sql_query, args)

        results2 = [(author, sid, path + pre + post, time) for (author, sid, path, pre, post, time) in cursor if sid not in sids]
        results.extend(results2)
                
        # filter out identical results
        sids = []
        results2 = []
        for result in results:
            (author, sid, text, time) = result

            if sid not in sids:
                results2.append(result)
                sids.append(sid)

        results = results2
        
        for author, sid, text, time in results:
            yield (self.env.href.shellsession(sid),
                    'Shell Log %s' % sid,
                    time, author,
                    shorten_result(text, terms))


    # INavigationContributor methods

    def get_active_navigation_item(self, req):
        return 'shellsession'

    def get_navigation_items(self, req):
        if not req.perm.has_permission('SHELLLOGS_VIEW'):
            return
        yield ('mainnav', 'shellsession',
               Markup('<a href="%s" accesskey="2">Shell Sessions</a>',
                      self.env.href.shellsession()))

    # IPermissionRequestor methods

    def get_permission_actions(self):
        return ['SHELLLOGS_VIEW']

    # IRequestHandler methods

    def match_request(self, req):
        match = re.match(r'^/shellsession(?:/(.*))?', req.path_info)
        if match:
            if match.group(1):
                req.args['page'] = match.group(1)
            return 1

    def process_request(self, req):
        req.perm.assert_permission('SHELLLOGS_VIEW')
        action = req.args.get('action', 'view')
        logname = req.args.get('page', 'SessionIndex')
        version = req.args.get('version')
        db = self.env.get_db_cnx()

        add_stylesheet(req, 'common/css/wiki.css')
        
        if logname.endswith('/'):
            req.redirect(self.env.href.wiki(logname.strip('/')))
        
        match = re.match('(\d+)-(\d+)', logname)
        if match:
           ticket = int(match.group(1))
           seq = int(match.group(2))
           req.hdf['shellsession.ticket'] = ticket
           req.hdf['shellsession.seq'] = seq

        if not match and not logname == 'SessionIndex':
            req.redirect(self.env.href.shellsession('SessionIndex'))



                    
        if logname == 'SessionIndex':
            req.hdf['shellsession.index'] = 1

            logs = self.getShellLogs()

            text = ''
            tkt = None
            for log in logs:
                sid, starttime = log
                tid, seq = sid.split('-')

                if tkt is None or tkt.id != int(tid):
                    tkt = Ticket(self.env, int(tid), db)
                    text += '\n * Ticket #%s: "%s"\n' % (tkt.id, tkt['summary'])

                text += '   * [shellsession:%s %s]\n' % (sid, ctime(starttime))

            req.hdf['shellsession.text'] = wiki_to_html(text, self.env, req)

        else:
            log = ShellLog(self.env, ticket, seq, version, db)
            if None == log.text:
                req.redirect(self.env.href.shellsession('SessionIndex'))

            if 'POST' == req.method:
                if 'edit' == action:
                    latest_version = ShellLog(self.env, ticket, seq, None, db).version
                    if req.args.has_key('cancel'):
                        req.redirect(self.env.href.shellsession(logname))
                    elif int(version) != latest_version:
                        action = 'collision'
                        self._render_editor(req, db, log)
                    else:
                        self._do_save(req, db, log)
                        

            elif 'context' == action:
                # increase or decrease the context lines for
                # a particular diff by 5 at a time

                key = req.args.get('key')
                change = req.args.get('change')
                self.env.log.debug("got context change (%s, %s)" % (key, change))

                sessionkey = 'tmp.shellsession.%s-%s.filediff.%s.context' % (ticket, seq, key)
                val = 5
                if sessionkey in req.session:
                    val = int(req.session[sessionkey])

                if change == 'more':
                    val += 5
                else:
                    val -=5

                val = max(0, val)

                req.session[sessionkey] = val

                req.redirect(self.env.href.shellsession('%s-%s' % (ticket, seq)))

                
            elif 'view' == action:
                req.hdf['shellsession.index'] = 0
                req.hdf['shellsession.text'] = wiki_to_html(log.text, self.env, req)

                diffs = self._getFilesForSession(ticket, seq)

                i = 0
                for q in diffs:

                    diff = FileDiff(self.env, ticket, seq, q)

                    if 1 == diff.active:
                        interp = diff.interpret()

                        req.hdf['shellsession.diff.%d.path' % i] = q
                        req.hdf['shellsession.diff.%d.change' % i] = interp

                        # the sessionkey for this file
                        # it will be deleted at logout since it begins with 'tmp.'
                        path_md5 = md5.new(diff.path).hexdigest()
                        sessionkey = 'tmp.shellsession.%s-%s.filediff.%s' % (ticket, seq, path_md5)
                        req.hdf['shellsession.diff.%d.sessionkey' % i] = path_md5

                        # retrieve the number of context lines from
                        # the session if it exists, else use the default (5)
                        context_key = '%s.context' % sessionkey
                        if not context_key in req.session:
                            context = 5
                        else:
                            context = int(req.session[context_key])

                        req.hdf['shellsession.diff.%d.context' % i] = context

                        self.env.log.debug(' diff on %s' % q)
                        if interp == 'modified':
                            self.env.log.debug('  => changed')
                            req.hdf['shellsession.diff.%d.diff' % i] = hdf_diff(
                                    diff.start.splitlines(), 
                                    diff.finish.splitlines(),
                                    context = context)

                        elif interp == 'added':
                            self.env.log.debug('  => added')
                            req.hdf['shellsession.diff.%d.diff' % i] = hdf_diff([], diff.finish.splitlines())
                            req.hdf['shellsession.diff.%d.diff' % i] = hdf_diff(
                                    [],
                                    diff.finish.splitlines(),
                                    context = context)

                        elif interp == 'deleted':
                            self.env.log.debug('  => deleted')
                            req.hdf['shellsession.diff.%d.diff' % i] = hdf_diff(
                                    diff.start.splitlines(), 
                                    [],
                                    context = context)

                        elif interp == 'temporary':
                            self.env.log.debug('  => tmp')

                        else:
                            self.env.log.debug('  => same')
                            req.hdf['shellsession.diff.%d.diff' % i] = []

                        i += 1

                req.hdf['diff.style'] = 'inline'
            
            elif action == 'edit':
                self._render_editor(req, db, log)
        
        req.hdf['wiki.action'] = action
        
        add_stylesheet(req, 'common/css/changeset.css')
        add_stylesheet(req, 'common/css/diff.css')
        add_stylesheet(req, 'common/css/code.css')
        
        return 'shellsession.cs', None

    # Internal Methods

    def _render_editor(self, req, db, log):
        logname = '%s-%s' % (log.ticket, log.seq)

        if req.args.has_key('text'):
            log.text = req.args.get('text')

        author = req.args.get('author', get_reporter_id(req))
        editrows = req.args.get('editrows')
        if editrows:
            pref = req.session.get('wiki_editrows', '20')
            if editrows != pref:
                req.session['wiki_editrows'] = editrows
        else:
            editrows = req.session.get('wiki_editrows', '20')
        
        req.hdf['title'] = logname
        info = {
            'page_name': logname,
            'page_source': log.text,
            'version': log.version,
            'author': author,
            'edit_rows': editrows,
            'scroll_bar_pos': req.args.get('scroll_bar_pos', '')
        }
       
        diffs = self._getFilesForSession(log.ticket, log.seq)
        i = 0
        for q in diffs:
            diff = FileDiff(self.env, log.ticket, log.seq, q)
            req.hdf['shellsession.diff.%d.path' % i] = q
            if 1 == diff.active:
                req.hdf['shellsession.diff.%d.active' % i] = 1
            else:
                req.hdf['shellsession.diff.%d.active' % i] = 0
            i += 1

        req.hdf['wiki'] = info
    
    def _do_save(self, req, db, log):
        log.old_text = log.text
        log.text = req.args.get('text')
        
        logname = '%s-%s' % (log.ticket, log.seq)

        diffs = self._getFilesForSession(log.ticket, log.seq)
        i = 0
        for q in diffs:
            diff = FileDiff(self.env, log.ticket, log.seq, q)
            diff.active = req.args.get('diff-%s' % diff.path) == 'show'
            diff.updateActive(req.remote_addr)
        if log.text != log.old_text:
            log.save(req.args.get('author'), req.remote_addr)
        req.redirect(self.env.href.shellsession(logname))


    def _getFilesForSession(self, ticket, seq):
        """
        Get all diff file names associated with a particular session ID

        @param ticket: the ticket id of the shell session
        @type  ticket: int

        @param seq: the shell session sequence number
        @type  seq: int

        @return: a list of filenames
        @rtype: list of strings
        """

        db = self.env.get_db_cnx()
        cursor = db.cursor()

        cursor.execute("SELECT path FROM filediff WHERE ticket=%d AND seq=%d ", (ticket, seq))

        rows = cursor.fetchall()
        
        filenames = []
        
        if rows:
            for row in rows:
                filename = row[0]
                filenames.append(filename)

        return filenames
    
    def getShellLogs(self):
        """
        Get all sessionIDs

        @return: a list of sessionIDs
        @rtype: list of strings
        """

        db = self.env.get_db_cnx()
        cursor = db.cursor()

        # cursor.execute("SELECT DISTINCT ticket,seq,start FROM shellsession ORDER BY ticket ASC, seq DESC")
        cursor.execute("""SELECT DISTINCT 
                                 d.ticket
                               , s.seq
                               , s.finish as finish_time
                               , ( CASE WHEN t.status = 'closed' THEN 1 
                                                                 ELSE 0 END ) as ticket_is_closed
                               , d.max_ticket_time

                            /* subquery gets the max time of a shell session grouped
                               by ticket; this helps us get the order we want */
                            FROM ( SELECT ticket
                                        , max(finish) as max_ticket_time

                                     FROM shellsession d

                                    GROUP BY ticket
                                    ORDER BY max_ticket_time DESC
                                 ) as d
                               , shellsession s
                               , ticket t
                               
                           WHERE s.ticket = d.ticket
                             AND t.id = s.ticket
                             
                           ORDER BY ticket_is_closed asc
                                  , max_ticket_time desc
                                  , finish_time desc
                       """)

        rows = cursor.fetchall()

        sessionIDs = []

        if rows:
            for (ticket, seq, finish_time, ticket_is_closed, max_ticket_time) in rows:
                sessionIDs.append(('%s-%s' % (ticket, seq), finish_time))

        return sessionIDs
