import os, os.path, shutil, socket
import time
import stat
import re

import trackle.wiki.model
from trackle.shelllogs.model import FileDiff, ShellLog
from trackle.shelllogs.api import statToDict, interpretStats

# for printing at STS prompt
from trackle import __version__ as TRACKLE_VERSION

def fileIn(path, dirs):
    """
    returns true iff the first n chars of
    path are equal to any of the dirs (a tuple)
    for n = len(thing in dirs).
    """
    for dir in dirs:
        l = len(dir)

        if path[:l] == dir:
            return True

    return False


class TrackleNoPermException(Exception):
    pass

class ShellTrackingSession:
    """
    This class maintains state of the Shell Tracking Session and
    controls starting and stopping it. It is called from within the
    Trackle CLI, after returning to shell mode (from curses terminal
    mode).

    The important functions are run() and saveConfig().

    @author Matt Singleton <msingle1@sccs.swarthmore.edu>
    @author Dan Crosta <dcrosta@sccs.swarthmore.edu>
    """

    def __init__(self, ticketID, env, ticket, user):
        self.env = env

        self.ticketID = ticketID
        self.tempDir = self.env.config.get('trackle', 'tempdir')
        self.libDir  = self.env.config.get('trackle', 'libdir')
        self.ticket = ticket
        self.ticket = ticket
        self.username = user
        
        # session id appends time of session start to ticket id
        # sessionDir lives in tempDir/session/sessionID
        # filesDir lives in sessionDir
        self.sessionID = '%d-%d' % (self.ticketID, time.time())
        self.sessionDir = os.path.join(self.tempDir, 'sessions-%s' % os.environ['USER'], self.sessionID)
        self.filesDir = os.path.join(self.sessionDir, 'files')
        
        #set up dirs for logging
        try:
            os.makedirs(self.sessionDir)
        except (IOError, OSError):
            raise TrackleNoPermException, self.sessionDir
        try:
            os.mkdir(self.filesDir)
        except (IOError, OSError):
            raise TrackleNoPermException, self.filesDir

    def makeWikiPage(self, fileList):
        ip = socket.gethostbyname(socket.gethostname())

        
        slog = ShellLog(self.env, self.ticketID)
 
        slog.start = self.startTime
        slog.finish = self.endTime

       
        # add diffs to page
        indexfile = open(os.path.join(self.sessionDir, 'index'), 'r')
        index = indexfile.readlines()
        indexfile.close()

        for s in index:
            s = s.strip()
            (source, filenum) = s.split(',')
            source = os.path.normpath(source)

            if source not in self.allFiles:
                continue

            fdiff = FileDiff(self.env, self.ticketID, slog.seq, source, new = True)

            fp = file(os.path.join( self.sessionDir, 'files', '%s.stat' % filenum ), 'r')
            fdiff.sstat = fp.readlines()
            fp.close()

            fp = file(os.path.join( self.sessionDir, 'files', '%s.stat.end' % filenum ), 'r')
            fdiff.fstat = fp.readlines()
            fp.close()


            sstatd = statToDict(fdiff.sstat)
            fstatd = statToDict(fdiff.fstat)


            # make them into strings for saving to the db
            fdiff.sstat = ''.join(fdiff.sstat)
            fdiff.fstat = ''.join(fdiff.fstat)

            if 'special' not in sstatd:
                self.env.log.debug("** missing special in sstat")
                self.env.log.debug("   %s" % source)
                self.env.log.debug("   %s" % sstatd)
            if 'special' not in fstatd:
                self.env.log.debug("** missing special in fstat")
                self.env.log.debug("   %s" % source)
                self.env.log.debug("   %s" % fstatd)



            fdiff.start = None
            fdiff.finish = None

            if 'noexist' not in sstatd['special']:
                fp = file(os.path.join(self.sessionDir, 'files', '%s.file' % filenum), 'r')
                fdiff.start = fp.read()
                fp.close()

            if 'noexist' not in fstatd['special']:
                fp = file(os.path.join(self.sessionDir, 'files', '%s.file.end' % filenum), 'r')
                fdiff.finish = fp.read()
                fp.close()


            if source in fileList:
                fdiff.active = 1

            else:
                fdiff.active = 0

            fdiff.save()


        # create the default wiki text
        fp = file(os.path.join(self.sessionDir, 'env'), 'r')
        env_vars = fp.read().strip()
        fp.close()

        fp = file(os.path.join(self.sessionDir, 'hist.bash'), 'r')
        shell_hist = fp.read().strip()
        fp.close()

        slog.text = """
(Jump to [ticket:%(ticket)s ticket])

== Start Time ==
%(stime)s

== End Time ==
%(ftime)s

== Environment Variables ==
{{{
%(env)s
}}}

== Shell History ==
{{{
%(hist)s
}}}

""" % {
    'ticket':slog.ticket,
    'seq'   :slog.seq,
    'stime' :time.ctime(self.startTime),
    'ftime' :time.ctime(self.endTime),
    'env'   :env_vars,
    'hist'  :shell_hist
    }
                       

        slog.save(self.username, ip)

        self.sessionID = '%d-%d' % (self.ticketID, slog.seq)

        return self.sessionID


    def run(self):
        """
        actually run the shell tracking session
        """
        
        # Store some shell info
        self.startTime = int(time.time())
        fp = file( os.path.join(self.sessionDir, 'time.start'), 'w')
        fp.write('%d\n' % self.startTime)
        fp.close()

        envS = '\n'.join(['%s = %s' % (s, os.environ[s]) for s in os.environ])
        fp = file( os.path.join(self.sessionDir, 'env'), 'w')
        fp.write(envS)
        fp.write('\n')
        fp.close()

        # not sure we can get this in python...
        os.system('echo `alias` > %s' % os.path.join(self.sessionDir, 'alias'))
        
        shell = os.environ['SHELL']
        oldhistfile = '' 
        
        # Set BASH as the shell and set the session history log file
        os.environ['SHELL'] = '/bin/bash'
        os.mknod(os.path.join(self.sessionDir, 'hist.bash'))
        if 'HISTFILE' in os.environ:
            oldhistfile = os.environ['HISTFILE']
        os.environ['HISTFILE'] = os.path.join(self.sessionDir, 'hist.bash')
        
        # Set some trackle-specific variables
        os.environ['TRACKLE_SESSION_DIR'] = self.sessionDir
        os.environ['TRACKLE_FILE_COUNTER'] = os.path.join(self.sessionDir, 'counter')
        os.environ['TRACKLE_UID'] = str(os.getuid())

        os.mknod(os.path.join(self.sessionDir, 'index'))
        os.mknod(os.path.join(self.sessionDir, 'counter'))

        # Set up library interposition to catch system calls
        os.environ['LD_PRELOAD'] = os.path.join(self.libDir, 'libtrackle.so')

        # Print a friendly reminder
        os.system('echo "%s"' % ("""

Welcome to Trackle STS!
-----------------------
* Your shell has been changed to BASH
* Everything you do will be logged
* Don't mess with environment variables
* Fix something
* Report bugs at https://gyr.sccs.swarthmore.edu/src
  (this is Trackle version %(version)s)
* Type 'exit' when finished

""" % {'version':TRACKLE_VERSION}).replace('"', '\"'))




        # Start the session
        os.system('script -c "bash --login" -q %s -t 2> %s' % (os.path.join(self.sessionDir, 'typescript'), os.path.join(self.sessionDir, 'timingfile')))

        # grab a copy of the shell history to pass
        # back to the CLI
        histfp = file(os.environ['HISTFILE'], 'r')
        history = histfp.readlines()
        histfp.close()

        # Clean up after trackle
        del os.environ['LD_PRELOAD']
        del os.environ['TRACKLE_SESSION_DIR']
        del os.environ['TRACKLE_FILE_COUNTER']
        del os.environ['TRACKLE_UID']
        if oldhistfile != '':
            os.environ['HISTFILE'] = oldhistfile
        else:
            del os.environ['HISTFILE']
        os.environ['SHELL'] = shell


        self.endTime = int(time.time())
        fp = file( os.path.join(self.sessionDir, 'time.end'), 'w' )
        fp.write('%d\n' % self.endTime)
        fp.close()
       
        # get the end version of all accessible files
        indexfile = open(os.path.join(self.sessionDir, 'index'), 'r')
        index = indexfile.readlines()
        indexfile.close()

        fileList = []
        for s in index:
            s = s.strip()
            (source, filenum) = s.split(',')
            source = os.path.normpath(source)

            # skip files in these places
            if fileIn(source, ('/proc', '/dev', '/sys', self.sessionDir)):
                continue

            # find out what happened to the file
            status = self.interpretStatFiles(filenum)

            if status == 'modified':
                # test to see if the file is binary -- if so, don't include it in
                # the shell session log (not even as a candidate)
                diffpipe = os.popen('diff %s %s' % (
                    os.path.join(self.sessionDir, 'files/%s.file' % filenum),
                    os.path.join(self.sessionDir, 'files/%s.file.end' % filenum)))
                diffline = diffpipe.readline().strip()
                diffpipe.close()

                # ignore files that are binary
                if re.match('Binary files.*differ$', diffline):
                    continue

            if status in ('modified', 'added', 'deleted', 'temporary'):
                fileList.append(source)


        self.allFiles = fileList
       
        return self.sessionID, fileList, history


    def cleanup(self, path = None):
        """
        recursively clean up by removing the session dir and all files in it
        """
        if path is None:
            path = self.sessionDir

        for p in os.listdir(path):
            p = os.path.join(path, p)
            if os.path.isdir(p):
                self.cleanup(p)
            else:
                os.remove(p)

        os.rmdir(path)

    def interpretStatFiles(self, filenum):
        """
        looks at the .stat and .stat.end files, and interprets
        what happened to the file durint the session. will return:
        'same', 'added', 'deleted' or 'changed'. a status of 'changed'
        implies that the <num>.file and <num>.file.end files will
        exist, but this is not verified
        """

        fp = file(os.path.join( self.sessionDir, 'files', '%s.stat' % filenum ), 'r')
        sstat = fp.readlines()
        fp.close()

        fp = file(os.path.join( self.sessionDir, 'files', '%s.stat.end' % filenum ), 'r')
        fstat = fp.readlines()
        fp.close()

        # translate them to dicts
        sstat = statToDict(sstat)
        fstat = statToDict(fstat)

        return interpretStats(sstat, fstat)
