# -*- coding: iso-8859-1 -*-
#
# Copyright (C) 2003-2006 Edgewall Software
# Copyright (C) 2003-2005 Jonas Borgstrm <jonas@edgewall.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Jonas Borgstrm <jonas@edgewall.com>

import re

from trackle import  util
from trackle.core import  *
from trackle.perm import  IPermissionRequestor
from trackle.wiki import  IWikiSyntaxProvider, Formatter
from trackle.Search import  ISearchSource, search_to_sql, shorten_result


class TicketSystem(Component):
    implements(IPermissionRequestor, IWikiSyntaxProvider, ISearchSource)

    # Public API

    def get_available_actions(self, ticket, perm_):
        """Returns the actions that can be performed on the ticket."""
        actions = {
            'new':         ['leave', 'resolve', 'reassign', 'accept', 'confirmcc', 'addcc'],
            'assigned':    ['leave', 'resolve', 'reassign',           'confirmcc', 'addcc'],
            'reopened':    ['leave', 'resolve', 'reassign',           'confirmcc', 'addcc'],
            'closed':      ['leave',                        'reopen'],
            'unconfirmed': ['leave', 'resolve', 'reassign', 'accept', 'confirm', 'confirmcc', 'addcc']
        }
        perms = {'resolve': 'TICKET_MODIFY', 'reassign': 'TICKET_CHGPROP',
                 'accept': 'TICKET_CHGPROP', 'reopen': 'TICKET_CREATE',
                 'conrirm': 'TICKET_CHGPROP', 'addcc': 'TICKET_ADD_CC',
                 'confirmcc': 'TICKET_ADD_CC'}
        return [action for action in actions.get(ticket['status'], ['leave'])
                if action not in perms or perm_.has_permission(perms[action])]

    def get_ticket_fields(self):
        """Returns the list of fields available for tickets."""
        from trackle.ticket import  model

        db = self.env.get_db_cnx()
        fields = []

        # Basic text fields
        for name in ('summary', 'reporter'):
            field = {'name': name, 'type': 'text', 'label': name.title()}
            fields.append(field)

        # Owner field, can be text or drop-down depending on configuration
        field = {'name': 'owner', 'label': 'Owner'}
        if self.config.getbool('ticket', 'restrict_owner'):
            field['type'] = 'select'
            users = []
            for username, name, email in self.env.get_known_users(db):
                users.append(username)
            field['options'] = users
            field['optional'] = True
        else:
            field['type'] = 'text'
        fields.append(field)

        # Description
        fields.append({'name': 'description', 'type': 'textarea',
                       'label': 'Description'})

        # Default select and radio fields
        selects = [('type', model.Type), ('status', model.Status),
                   ('priority', model.Priority),
                   ('severity', model.Severity),
                   ('milestone', model.Milestone),
                   ('component', model.Component),
                   ('resolution', model.Resolution)]
        for name, cls in selects:
            options = [val.name for val in cls.select(self.env, db=db)]
            if not options:
                # Fields without possible values are treated as if they didn't
                # exist
                continue
            field = {'name': name, 'type': 'select', 'label': name.title(),
                     'value': self.config.get('ticket', 'default_' + name),
                     'options': options}

            if name is 'severity':
                field['label'] = 'User Severity'
                field['optional'] = True

            if name in ('priority', 'milestone', 'component', 'version'):
                field['optional'] = True
            elif name in ('status', 'resolution'):
                field['type'] = 'radio'

            fields.append(field)

        # Advanced text fields
        for name in ('keywords', 'cc', ):
            field = {'name': name, 'type': 'text', 'label': name.title()}
            fields.append(field)

        fields.append({'name': 'locked', 'type': 'select', 'label': 'Locked', 'value': 'False', 'options': ['False'], 'skip': True})
        fields.append({'name': 'public', 'type': 'checkbox', 'label': 'Public', 'value':1, 'skiptop':True})

        for field in self.get_custom_fields():
            if field['name'] in [f['name'] for f in fields]:
                self.log.warning('Duplicate field name "%s" (ignoring)',
                                 field['name'])
                continue
            if not re.match('^[a-zA-Z][a-zA-Z0-9_]+$', field['name']):
                self.log.warning('Invalid name for custom field: "%s" '
                                 '(ignoring)', field['name'])
                continue
            field['custom'] = True
            fields.append(field)

        return fields

    def get_custom_fields(self):
        fields = []
        for name in [option for option, value
                     in self.config.options('ticket-custom')
                     if '.' not in option]:
            field = {
                'name': name,
                'type': self.config.get('ticket-custom', name),
                'order': int(self.config.get('ticket-custom', name + '.order', '0')),
                'label': self.config.get('ticket-custom', name + '.label') \
                         or name.capitalize(),
                'value': self.config.get('ticket-custom', name + '.value', '')
            }
            if field['type'] == 'select' or field['type'] == 'radio':
                options = self.config.get('ticket-custom', name + '.options')
                field['options'] = [value.strip() for value in options.split('|')]
            elif field['type'] == 'textarea':
                field['width'] = self.config.get('ticket-custom', name + '.cols')
                field['height'] = self.config.get('ticket-custom', name + '.rows')
            fields.append(field)

        fields.sort(lambda x, y: cmp(x['order'], y['order']))
        return fields

    def get_user_field_names(self):
        userfields = self.env.config.get('trackle', 'user_fields')
        userfields = [field.strip() for field in userfields.replace(',', ' ').split()]
        userfields.extend(['summary', 'reporter', 'description', 'type', 'status', 'resolution'])

        return userfields
    

    # IPermissionRequestor methods

    def get_permission_actions(self):
        return ['TICKET_APPEND', 'TICKET_CREATE', 'TICKET_CHGPROP',
                'TICKET_VIEW_BASIC', 'TICKET_VIEW_EXTENDED', 'TICKET_ADD_CC',
                ('TICKET_MODIFY', ['TICKET_APPEND', 'TICKET_CHGPROP',
                                   'TICKET_ADD_CC']),  
                ('TICKET_ADMIN', ['TICKET_CREATE', 'TICKET_MODIFY',  
                                  'TICKET_VIEW_BASIC', 'TICKET_VIEW_EXTENDED',
                                  'TICKET_ADD_CC'])]

    # IWikiSyntaxProvider methods

    def get_link_resolvers(self):
        return [('bug', self._format_link),
                ('ticket', self._format_link)]

    def get_wiki_syntax(self):
        yield (
            # matches #... but not &#... (HTML entity)
            r"!?(?<!&)#"
            # optional intertrac shorthand #T... + digits
            r"(?P<it_ticket>%s)?\d+" % Formatter.INTERTRAC_SCHEME,
            lambda x, y, z: self._format_link(x, 'ticket', y[1:], y, z))
        yield (
            r"\[newticket(?: (?P<lbl>[A-Za-z0-9_ ]*))?\]",
            lambda formatter, ns, match: self._format_link(formatter, 'newticket', None, match.group('label'), match))

    def _format_link(self, formatter, ns, target, label, fullmatch=None):
        if ns == 'newticket':
            if fullmatch.group('lbl') is not None:
                label = fullmatch.group('lbl')

            return '<a href="%(href)s" title="%(msg)s">%(msg)s</a>' % {
                'href': formatter.href.newticket(),
                'msg' : label or 'Create a New Ticket' }

        else:
            intertrac = formatter.shorthand_intertrac_helper(ns, target, label,
                                                             fullmatch)
            if intertrac:
                return intertrac

            cursor = formatter.db.cursor()
            cursor.execute("SELECT summary,status FROM ticket WHERE id=%s",
                           (target,))
            row = cursor.fetchone()
            if row:
                summary = util.escape(util.shorten_line(row[0]))
                return '<a class="%s ticket" href="%s" title="%s (%s)">%s</a>' \
                       % (row[1], formatter.href.ticket(target), summary, row[1],
                          label)
            else:
                return '<a class="missing ticket" href="%s" rel="nofollow">%s</a>' \
                       % (formatter.href.ticket(target), label)

    
    # ISearchProvider methods

    def get_search_filters(self, req):
        if req.perm.has_permission('TICKET_VIEW_BASIC'):
            yield ('ticket', 'Tickets')

    def get_search_results(self, req, terms, filters):
        if not 'ticket' in filters:
            return
        db = self.env.get_db_cnx()

        if req.authname == 'anonymous':
            # anonymous users are only allowed to search 
            # some of the fields, and none of the change
            # history, and only search confirmed tickets
            sql, args = search_to_sql(db, ['summary', 'keywords', 'description'], terms)
            cursor = db.cursor()
            cursor.execute("SELECT DISTINCT summary, description, '[email hidden]', "
                           "keywords, id, time FROM ticket "
                           "WHERE status != 'unconfirmed' AND public = 1 AND (%s)" % sql,
                           args)
            
        else:
            sql, args = search_to_sql(db, ['b.newvalue'], terms)
            sql2, args2 = search_to_sql(db, ['summary', 'keywords', 'description', 'reporter', 'cc'], terms)
            cursor = db.cursor()
            cursor.execute("SELECT DISTINCT a.summary,a.description,a.reporter, "
                           "a.keywords,a.id,a.time FROM ticket a "
                           "LEFT JOIN ticket_change b ON a.id = b.ticket "
                           "WHERE (b.field='comment' AND %s ) OR %s" % (sql, sql2),
                           args + args2)

        for summary,desc,author,keywords,tid,date in cursor:
            yield (self.env.href.ticket(tid),
                   '#%d: %s' % (tid, util.shorten_line(summary)),
                   date, author,
                   shorten_result(desc, terms))
