# -*- coding: iso-8859-1 -*-
#
# Copyright (C) 2003-2006 Edgewall Software
# Copyright (C) 2003-2005 Daniel Lundin <daniel@edgewall.com>
# Copyright (C) 2005-2006 Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Daniel Lundin <daniel@edgewall.com>
#

from trackle import  __version__
from trackle.core import  TracError
from trackle.util import  CRLF, wrap
from trackle.notification import  NotifyEmail

import md5
import time
import base64

class TicketNotifyEmail(NotifyEmail):
    """Notification of ticket changes."""

    template_name = "ticket_notify_email.cs"
    ticket = None
    newticket = None
    modtime = 0
    from_email = 'trac+ticket@localhost'
    COLS = 75

    # in most cases, we want to send the limited version
    # of the email which contains fewer fields
    limited = True

    def __init__(self, env):
        NotifyEmail.__init__(self, env)
        self.prev_cc = []

        # these are the fields the user is allowed to see changes to
        self.limited_fields = self.config.get('trackle', 'user_fields')
        self.limited_fields = self.limited_fields.replace(',', ' ').split()

        # make sure owner and status are in there
        if 'owner' not in self.limited_fields:
            self.limited_fields.append('owner')
        if 'status' not in self.limited_fields:
            self.limited_fields.append('status')
        if 'resolution' not in self.limited_fields:
            # things like 'fixed', 'invalid', etc
            self.limited_fields.append('resolution')
        

    def notify(self, ticket, newticket=True, modtime=0, allowedfields=[]):
        self.ticket = ticket
        self.modtime = modtime
        self.newticket = newticket

        if self.ticket['status'] == 'unconfirmed':
            self.hdf['email.ticket.unconfirmed'] = 'true'
            self.hdf['email.ticket.confirmurl'] = self.env.abs_href.ticket(self.ticket.id, 'confirm', self.ticket.get_confirm_key())

        self.ticket['description'] = wrap(self.ticket.values.get('description', ''),
                                          self.COLS, initial_indent=' ',
                                          subsequent_indent=' ', linesep=CRLF)
        self.ticket['link'] = self.env.abs_href.ticket(ticket.id)
        self.hdf.set_unescaped('email.ticket_props', self.format_props())
        self.hdf.set_unescaped('email.ticket_body_hdr', self.format_hdr())
        self.hdf.set_unescaped('ticket', self.ticket.values)
        self.hdf['ticket.new'] = self.newticket
        self.hdf['ticket.id'] = ticket.id
        subject = self.format_subj()
        if not self.newticket:
            subject = 'Re: ' + subject
        self.hdf.set_unescaped('email.subject', subject)
        changes = ''
         
        for field in allowedfields:
            if field not in self.limited_fields:
                self.limited_fields.append(field)

        if not self.newticket and modtime:  # Ticket change
            changelog = ticket.get_changelog(modtime)
            for date, author, field, old, new in changelog:
                if self.limited and field not in self.limited_fields:
                    # only show changes to fields the user is allowed to see
                    continue

                self.hdf.set_unescaped('ticket.change.author', author)
                pfx = 'ticket.change.%s' % field
                newv = ''
                if field == 'comment':
                    newv = wrap(new, self.COLS, ' ', ' ', CRLF)
                elif field == 'description':
                    new_descr = wrap(new, self.COLS, ' ', ' ', CRLF)
                    old_descr = wrap(old, self.COLS, '> ', '> ', CRLF)
                    old_descr = old_descr.replace(2*CRLF, CRLF + '>' + CRLF)
                    cdescr = CRLF
                    cdescr += 'Old description:' + 2*CRLF + old_descr + 2*CRLF
                    cdescr += 'New description:' + 2*CRLF + new_descr + CRLF
                    self.hdf.set_unescaped('email.changes_descr', cdescr)
                else:
                    newv = new
                    l = 7 + len(field)
                    chg = wrap('%s => %s' % (old, new), self.COLS - l, '',
                               l * ' ', CRLF)
                    changes += '  * %s:  %s%s' % (field, chg, CRLF)
                if newv:
                    self.hdf.set_unescaped('%s.oldvalue' % pfx, old)
                    self.hdf.set_unescaped('%s.newvalue' % pfx, newv)
                if field == 'cc':
                    self.prev_cc += old and self.parse_cc(old) or []
                self.hdf.set_unescaped('%s.author' % pfx, author)
            if changes:
                self.hdf.set_unescaped('email.changes_body', changes)
        NotifyEmail.notify(self, ticket.id, subject)

    def format_props(self):
        tkt = self.ticket
        fields = [f for f in tkt.fields if f['name'] not in ('summary', 'cc', 'locked')]
        if self.limited:
            # if limited, only show the fields that are allowed
            fields = [f for f in fields if f['name'] in self.limited_fields]

        # don't show fields with no value
        fields = [f for f in fields if self.ticket.values.has_key(f['name'])]

        width = [0, 0, 0, 0]
        i = 0
        for f in [f['name'] for f in fields if f['type'] != 'textarea']:
            if not tkt.values.has_key(f):
                continue
            fval = str(tkt[f])
            if fval.find('\n') != -1:
                continue
            idx = 2 * (i % 2)
            if len(f) > width[idx]:
                width[idx] = len(f)
            if len(fval) > width[idx + 1]:
                width[idx + 1] = len(fval)
            i += 1
        format = ('%%%is:  %%-%is  |  ' % (width[0], width[1]),
                  ' %%%is:  %%-%is%s' % (width[2], width[3], CRLF))
        l = (width[0] + width[1] + 5)
        sep = l * '-' + '+' + (self.COLS - l) * '-'
        txt = sep + CRLF
        big = []
        i = 0
        for f in [f for f in fields if f['name'] != 'description']:
            fname = f['name']
            if not tkt.values.has_key(fname):
                continue
            fval = tkt[fname]
            if f['type'] == 'textarea' or '\n' in str(fval):
                big.append((fname.capitalize(), CRLF.join(fval.splitlines())))
            else:
                txt += format[i % 2] % (fname.capitalize(), fval)
                i += 1
        if i % 2:
            txt += CRLF
        if big:
            txt += sep
            for name, value in big:
                txt += CRLF.join(['', name + ':', value, '', ''])
        txt += sep
        return txt

    def parse_cc(self, txt):
        return filter(lambda x: '@' in x, txt.replace(',', ' ').split())

    def format_hdr(self, prefix='Ticket '):
        return wrap("%s#%s: %s" % (
                    prefix,
                    self.ticket.id,
                    self.ticket['summary']),
                self.COLS, linesep = CRLF,
                subsequent_indent = ' '*(len(prefix) + len(str(self.ticket.id)) + 3)
                )

    def format_subj(self):
        projname = self.config.get('project', 'name')
        return '[%s] %s' % (projname, self.ticket['summary'])

    def get_recipients(self, tktid):
        notify_reporter = self.config.getbool('notification',
                                              'always_notify_reporter')
        notify_owner = self.config.getbool('notification',
                                           'always_notify_owner')

        ccrecipients = self.prev_cc
        torecipients = []
        cursor = self.db.cursor()

        # Harvest email addresses from the cc, reporter, and owner fields
        cursor.execute("SELECT cc,reporter,owner FROM ticket WHERE id=%s",
                       (tktid,))
        row = cursor.fetchone()
        if row:
            ccrecipients += row[0] and row[0].replace(',', ' ').split() or []
            if notify_reporter:
                torecipients.append(row[1])
            if notify_owner:
                torecipients.append(row[2])

        # Harvest email addresses from the author field of ticket_change(s)
        if notify_reporter:
            cursor.execute("SELECT DISTINCT author,ticket FROM ticket_change "
                           "WHERE ticket=%s", (tktid,))
            for author,ticket in cursor:
                torecipients.append(author)

        # Add smtp_always_cc address
        acc = self.config.get('notification', 'smtp_always_cc')
        if acc:
            ccrecipients += acc.replace(',', ' ').split()

        # now convert recipients into email addresses where necessary
        toemails = []
        ccemails = []
        for recipient in torecipients:
            if recipient.find('@') >= 0:
                toemails.append(recipient)
            elif self.email_map.has_key(recipient):
                toemails.append(self.email_map[recipient])
        for recipient in ccrecipients:
            if recipient.find('@') >= 0:
                ccemails.append(recipient)
            elif self.email_map.has_key(recipient):
                ccemails.append(self.email_map[recipient])

        # Remove duplicates
        toresult = []
        ccresult = []
        for e in toemails:
            if e not in toresult:
                toresult.append(e)
        for e in [c for c in ccemails if c not in toresult]:
            if e not in ccresult:
                ccresult.append(e)
        return (toresult, ccresult)

    def get_message_id(self, rcpt, modtime=0):
        """Generate a predictable, but sufficiently unique message ID."""
        s = '%s.%08d.%d.%s' % (self.config.get('project', 'url'),
                               int(self.ticket.id), modtime, rcpt)
        dig = md5.new(s).hexdigest()
        host = self.from_email[self.from_email.find('@') + 1:]
        msgid = '<%03d.%s@%s>' % (len(s), dig, host)
        return msgid

    def send(self, torcpts, ccrcpts):
        hdrs = {}
        dest = torcpts or ccrcpts
        if not dest:
            self.env.log.info('no recipient for a ticket notification')
            return 
        hdrs['Message-ID'] = self.get_message_id(dest[0], self.modtime)
        hdrs['X-Trac-Ticket-ID'] = str(self.ticket.id)
        hdrs['X-Trac-Ticket-URL'] = self.ticket['link']
        if not self.newticket:
            hdrs['In-Reply-To'] = self.get_message_id(dest[0])
            hdrs['References'] = self.get_message_id(dest[0])
        NotifyEmail.send(self, torcpts, ccrcpts, hdrs)


class TicketNotifyStaffEmail(TicketNotifyEmail):
    """
    this sub-class of TicketNotifyEmail only emails to the
    addresses listed in smtp_staff_addr in trackle.ini
    """

    def __init__(self, env):
        TicketNotifyEmail.__init__(self, env)

        self.limited = False

    def get_recipients(self, tktid):
        addresses = self.config.get('notification', 'smtp_staff_addr')

        # allow addresses to be separated by whitespace or commas
        return (addresses.replace(',', ' ').split(), [])


class PeriodicNotifyEmail(NotifyEmail, TicketNotifyEmail):

    template_name = 'ticket_periodic_email.cs'
    from_email = 'trac+notify@localhost'
    limited = False

    def __init__(self, env):
        # from Notify.__init__()
        from trackle.web.clearsilver import  HDFWrapper

        self.env = env
        self.config = env.config
        self.db = env.get_db_cnx()
        self.hdf = HDFWrapper(loadpaths=[env.get_templates_dir(),
                                         self.config.get('trac', 'templates_dir')])

        # from NotifyEmail.__init__()

        self._init_pref_encoding()
        # Get the email addresses of all known users
        self.email_map = {}
        for username, name, email in self.env.get_known_users(self.db):
            if email:
                self.email_map[username] = email
         

        self.now = time.time()



    def notify(self, open_tickets, ticket_changes, last_run_time):
        self.hdf.set_unescaped('email.tickets.open', self.format_open_tickets(open_tickets))
        self.hdf.set_unescaped('email.tickets.opendetail', self.format_open_tickets_detail(open_tickets))
        self.hdf.set_unescaped('email.tickets.changes', self.format_changes(ticket_changes))

        self.hdf['email.runtime'] = time.strftime('%m/%d/%y @ %I:%M %p').lower()
        self.hdf['email.lastruntime'] = time.strftime('%m/%d/%y @ %I:%M %p', time.localtime(last_run_time)).lower()

        NotifyEmail.notify(self, None, '[%s] Ticket Status Update' % self.config.get('project', 'name'))

    
    def format_open_tickets(self, tickets):
        """
        @param tickets: a list of Tickets
        """
        body = []

        for ticket in tickets:
            self.ticket = ticket
            txt = self.format_hdr(prefix = '- ')
            body.append(txt)

        return CRLF.join(body)


    def format_open_tickets_detail(self, tickets):
        """
        @param tickets: a list of Tickets
        """
        body = []

        for ticket in tickets:
            self.ticket = ticket
            txt = self.format_hdr()
            body.append(txt)

            txt = self.format_props()
            txt = CRLF.join(txt.split(CRLF)[:-1])
            txt += CRLF
            body.append(txt)

        return CRLF.join(body)
    

    def format_changes(self, tickets):
        """
        @param tickets: a list of TicketChangelog objects (see
            trackle/scripts/notify.py)
        """
        body = []

        for t in tickets:
            ticket  = t.ticket
            changes = t.changes

            self.ticket = ticket
            body.append(self.format_hdr())

            prevtime = None
            for change in changes:
                line = ''
                if change.time != prevtime:
                    when = self.prettytime(change.time),
                    line = '  %s: ' % when

                    prevtime = change.time

                else:
                    line = '  %s  ' % (' ' * 19)

                if change.action == 'closed':
                    line += 'closed (%s) by %s' % (change.value, change.author)

                elif change.action == 'reopened':
                    line += 'reopened by %s' % change.author

                elif change.action == 'assigned':
                    if change.value == change.author:
                        line += 'accepted by %s' % (change.author)
                    else:
                        line += 'assigned to %s by %s' % (change.value, change.author)

                body.append(line)

            body.append('')
        
        return CRLF.join(body)

        # NotifyEmail.notify(self, resid = None)

    def get_recipients(self, resid = None):
        addresses = self.config.get('notification', 'smtp_staff_addr')

        # allow addresses to be separated by whitespace or commas
        return (addresses.replace(',', ' ').split(), [])

    
    def prettytime(self, time1):
        delta = int(self.now - time1)
        time1 = time.localtime(time1)

        if delta < 3600 * 24 * 7:
            return 'Last %s @ %s' % (
                    time.strftime('%a', time1),
                    time.strftime('%I:%M %p', time1).lower())
        else:
            return time.strftime('%m/%d/%y @ %I:%M %p', time1).lower()


class TicketAddCcConfirmEmail(NotifyEmail):

    template_name = 'ticket_add_cc_email.cs'
    from_email = 'trac+notify@localhost'
    COLS = 75

    def notify(self, ticket, cc_email):
        self.added_cc = cc_email
        self.ticket = ticket

        encoded = base64.encodestring(self.added_cc).strip()
        self.hdf['email.ticket.confirmurl'] = self.env.abs_href.ticket(self.ticket.id, 'watch', encoded)

        self.hdf.set_unescaped('email.body.wrapped', wrap(
                'Someone has submitted this email address to be added as a '
                '"watcher" for %(projname)s ticket #%(id)s.'
                'If this is in error, do nothing and you won\'t be emailed '
                'again. If you wish to be added to the "watcher" list for '
                'this ticket, you must visit the link below either by clicking '
                'it or copying it and pasting into your web browser.' %
                {'projname': self.config.get('project', 'name'),
                 'id'      : self.ticket.id}, self.COLS, linesep=CRLF))

        subject = '[%s] Add Watcher Request' % self.config.get('project', 'name')
        NotifyEmail.notify(self, ticket, subject)

    def get_recipients(self, resid = None):
        return ([self.added_cc], [])
