# -*- coding: iso-8859-1 -*-
#
# Copyright (C) 2003-2006 Edgewall Software
# Copyright (C) 2003-2005 Jonas Borgstrm <jonas@edgewall.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Jonas Borgstrm <jonas@edgewall.com>

import os
import re
import time
import base64

from trackle import  util
from trackle.attachment import  attachment_to_hdf, Attachment
from trackle.core import  *
from trackle.env import  IEnvironmentSetupParticipant
from trackle.ticket import  Milestone, Ticket, TicketSystem
from trackle.ticket.notification import  TicketNotifyEmail, TicketNotifyStaffEmail, TicketAddCcConfirmEmail
from trackle.Timeline import  ITimelineEventProvider
from trackle.web import  IRequestHandler
from trackle.web.chrome import  add_link, add_stylesheet, INavigationContributor
from trackle.wiki import  wiki_to_html, wiki_to_oneliner


class NewticketModule(Component):

    implements(IEnvironmentSetupParticipant, INavigationContributor,
               IRequestHandler)

    # IEnvironmentSetupParticipant methods

    def environment_created(self):
        """Create the `site_newticket.cs` template file in the environment."""
        if self.env.path:
            templates_dir = os.path.join(self.env.path, 'templates')
            if not os.path.exists(templates_dir):
                os.mkdir(templates_dir)
            template_name = os.path.join(templates_dir, 'site_newticket.cs')
            template_file = file(template_name, 'w')
            template_file.write("""<?cs
####################################################################
# New ticket prelude - Included directly above the new ticket form
?>
""")

    def environment_needs_upgrade(self, db):
        return False

    def upgrade_environment(self, db):
        pass

    # INavigationContributor methods

    def get_active_navigation_item(self, req):
        return 'newticket'

    def get_navigation_items(self, req):
        if not req.perm.has_permission('TICKET_CREATE'):
            return
        yield ('mainnav', 'newticket', 
               util.Markup('<a href="%s" accesskey="7">New Ticket</a>',
                           self.env.href.newticket()))

    # IRequestHandler methods

    def match_request(self, req):
        return re.match(r'/newticket/?', req.path_info) is not None

    def process_request(self, req):
        req.perm.assert_permission('TICKET_CREATE')

        db = self.env.get_db_cnx()

        # some simple validation
        if req.authname == 'anonymous' and req.args.has_key('reporter'):
            email = util.validate_email(req.args.get('reporter'))
            if email is None:
                req.args['preview'] = 'true'
                req.hdf['validation.email.invalid'] = 'true'

        if req.method == 'POST' and not req.args.has_key('preview'):
            self._do_create(req, db)

        ticket = Ticket(self.env, db=db)
        ticket.populate(req.args)
        ticket.values.setdefault('reporter', util.get_reporter_id(req))

        if ticket.values.has_key('description'):
            description = wiki_to_html(ticket['description'], self.env, req, db)
            req.hdf['newticket.description_preview'] = description

        req.hdf['title'] = 'New Ticket'
        req.hdf['newticket'] = ticket.values

        field_names = [field['name'] for field in ticket.fields
                       if not field.get('custom')]
        if 'owner' in field_names:
            curr_idx = field_names.index('owner')
            if 'cc' in field_names:
                insert_idx = field_names.index('cc')
            else:
                insert_idx = len(field_names)
            if curr_idx < insert_idx:
                ticket.fields.insert(insert_idx, ticket.fields[curr_idx])
                del ticket.fields[curr_idx]

        # transform the comma-separated list into a python list of strings
        userfields = self.env.config.get('trackle', 'user_fields')
        userfields = [field.strip() for field in userfields.replace(',', ' ').split()]

        userfields = TicketSystem(self.env).get_user_field_names()

        for field in ticket.fields:
            name = field['name']
            del field['name']
            if name in ('summary', 'reporter', 'description', 'type', 'status',
                        'resolution'):
                field['skip'] = True
            elif name == 'owner':
                field['label'] = 'Assign to'
            elif name == 'milestone':
                # Don't make completed milestones available for selection
                options = field['options'][:]
                for option in field['options']:
                    milestone = Milestone(self.env, option, db=db)
                    if milestone.is_completed:
                        options.remove(option)
                field['options'] = options

            # these fields *are* shown to users and will be given explanation
            if req.authname == 'anonymous' and name in userfields:
                field['explanation'] = self.env.config.get('trackle', 'explanation_%s' % name)

            # otherwise, we skip them only if the user is not authenticated
            elif req.authname == 'anonymous':
                field['skip'] = True

            req.hdf['newticket.fields.' + name] = field


        add_stylesheet(req, 'common/css/ticket.css')
        return 'newticket.cs', None

    # Internal methods

    def _do_create(self, req, db):
        if not req.args.get('summary'):
            raise TracError('Tickets must contain a summary.')

        ticket = Ticket(self.env, db=db)
        ticket.values.setdefault('reporter', util.get_reporter_id(req))
        ticket.populate(req.args)

        # add the reporter email to cc
        email = util.validate_email(ticket['reporter'])
        if email is not None:
            if 'cc' in ticket.values.keys():
                ticket['cc'] += ', ' + ticket['reporter']
            else:
                ticket['cc'] = ticket['reporter']

        ticket.insert(db=db)
        db.commit()

        # Notify
        try:
            if ticket['status'] != 'unconfirmed':
                # don't notify staff when ticket is unconfirmed
                tn = TicketNotifyStaffEmail(self.env)
                tn.notify(ticket, newticket=True)
        except Exception, e:
            self.log.exception("Failure sending notification on creation of "
                               "ticket #%s: %s" % (ticket.id, e))
        try:
            tn = TicketNotifyEmail(self.env)
            tn.notify(ticket, newticket=True)
        except Exception, e:
            self.log.exception("Failure sending notification on creation of "
                               "ticket #%s: %s" % (ticket.id, e))

        # if the user is anonymous, pop up a message saying the ticket
        # is unconfirmed and what they should do about it
        if req.authname == 'anonymous':
            raise TracMessage(['The ticket you just submitted is not yet '
                    'confirmed! To finish the job, you need to click the link '
                    'in the email you will receive shortly, or copy and paste '
                    'the url into a web browser. Once you have completed that '
                    'step, your ticket will be marked as "confirmed," staff '
                    'will be notified, and the problem will soon get fixed! ',
                    'Your ticket has been given id "<strong>%d</strong>". If you '
                    'need to contact staff regarding this ticket, you should '
                    'use this name to refer to your ticket. This information '
                    'is also in the confirmation email you have been sent.' 
                    % ticket.id],
                    title = 'Unconfirmed Ticket Submitted',
                    redirect_url = self.env.href.ticket(ticket.id),
                    redirect_delay = 15)

        # Redirect the user to the newly created ticket
        req.redirect(self.env.href.ticket(ticket.id))


class TicketModule(Component):

    implements(INavigationContributor, IRequestHandler, ITimelineEventProvider)

    # INavigationContributor methods

    def get_active_navigation_item(self, req):
        return 'tickets'

    def get_navigation_items(self, req):
        return []

    # IRequestHandler methods

    def match_request(self, req):
        match = re.match(r'/ticket/([0-9]+)(?:(?:/(confirm)/(.*))|(?:/(watch)/(.*)))?', req.path_info)
        if match:
            req.args['id'] = match.group(1)
            if match.group(2) is not None:
                req.args['confirmkey'] = match.group(3)
            if match.group(4) is not None:
                req.args['watcher_enc'] = match.group(5)

            return True

    def process_request(self, req):
        req.perm.assert_permission('TICKET_VIEW_BASIC')

        action = req.args.get('action', 'view')

        if not req.args.has_key('id'):
            req.redirect(self.env.href.wiki())

        db = self.env.get_db_cnx()
        id = int(req.args.get('id'))

        ticket = Ticket(self.env, id, db=db)

        # fail early if anonymous is trying to view a private ticket
        if not ticket['public'] and (not req.authname or req.authname == 'anonymous'):
            raise TracMessage('Only authenticated staff may view this ticket.',
                    title = 'Private Ticket',
                    redirect_url=self.env.href.query())


        reporter_id = util.get_reporter_id(req)

        if req.args.has_key('confirmkey'):
            if ticket['status'] == 'unconfirmed':
                confirmkey = ticket.get_confirm_key()
    
                if confirmkey == req.args.get('confirmkey'):
                    req.args['ts'] = ticket.time_changed
                    req.args['comment'] = 'Ticket confirmed by user'
                    req.args['action'] = 'confirm' # so that _do_save will update the status
    
                    self._do_save(req, db, ticket, overlook_permissions = True)

                else:
                    raise TracError('You supplied an invalid confirmation key. Perhaps '
                        'you did not copy the full URL from the confirmation email?',
                        'Invalid Confirm Key')
            else:
                raise TracError('You cannot confirm this ticket, since it is already confirmed.',
                        'Cannont Confirm Ticket')

        elif req.args.has_key('watcher_enc'):
            email = base64.decodestring(req.args.get('watcher_enc'))
            email = util.validate_email(email)

            if email is None:
                raise TracError("The email address you have entered doesn't "
                        "look like a valid email address. Are you sure you copied "
                        "the entire confirmation link? If this is in "
                        "error, please email staff to report this bug.",
                        'Invalid email')
            
            req.args['action'] = 'confirmcc'
            req.args['addcc'] = email
            req.args['ts'] = ticket.time_changed

            self._do_save(req, db, ticket)

        elif req.args.has_key('addcc'):
            if ticket.values.has_key('cc'):
                alreadywatching = ticket['cc'].replace(',', ' ').split()

                if req.args.get('addcc') in alreadywatching:
                    raise TracError('Whoops! That email address is already listed to be '
                            'notified of updates to this ticket. No action will be taken. ',
                            'Email already watching')



        if req.method == 'POST':
            if not req.args.has_key('preview'):
                self._do_save(req, db, ticket)
            else:
                # Use user supplied values
                ticket.populate(req.args)
                req.hdf['ticket.status'] = ticket['status']
                req.hdf['ticket.action'] = action
                req.hdf['ticket.ts'] = req.args.get('ts')
                req.hdf['ticket.reassign_owner'] = req.args.get('reassign_owner') \
                                                   or req.authname
                req.hdf['ticket.resolve_resolution'] = req.args.get('resolve_resolution')
                reporter_id = req.args.get('author')
                comment = req.args.get('comment')
                if comment:
                    req.hdf['ticket.comment'] = comment
                    # Wiki format a preview of comment
                    req.hdf['ticket.comment_preview'] = wiki_to_html(comment,
                                                                     self.env,
                                                                     req, db)
        else:
            req.hdf['ticket.reassign_owner'] = req.authname
            # Store a timestamp in order to detect "mid air collisions"
            req.hdf['ticket.ts'] = ticket.time_changed

        self._insert_ticket_data(req, db, ticket, reporter_id)

        # If the ticket is being shown in the context of a query, add
        # links to help navigate in the query result set
        if 'query_tickets' in req.session:
            tickets = req.session['query_tickets'].split()
            if str(id) in tickets:
                idx = tickets.index(str(ticket.id))
                if idx > 0:
                    add_link(req, 'first', self.env.href.ticket(tickets[0]),
                             'Ticket #%s' % tickets[0])
                    add_link(req, 'prev', self.env.href.ticket(tickets[idx - 1]),
                             'Ticket #%s' % tickets[idx - 1])
                if idx < len(tickets) - 1:
                    add_link(req, 'next', self.env.href.ticket(tickets[idx + 1]),
                             'Ticket #%s' % tickets[idx + 1])
                    add_link(req, 'last', self.env.href.ticket(tickets[-1]),
                             'Ticket #%s' % tickets[-1])
                add_link(req, 'up', req.session['query_href'])

        add_stylesheet(req, 'common/css/ticket.css')
        return 'ticket.cs', None

    # ITimelineEventProvider methods

    def get_timeline_filters(self, req):
        if req.perm.has_permission('TICKET_VIEW_BASIC'):
            yield ('ticket', 'Ticket changes')
            if self.config.getbool('timeline', 'ticket_show_details'):
                yield ('ticket_details', 'Ticket details', False)

    def get_timeline_events(self, req, start, stop, filters):
        format = req.args.get('format')

        status_map = {'new': ('newticket', 'created'),
                      'reopened': ('newticket', 'reopened'),
                      'closed': ('closedticket', 'closed'),
                      'edit': ('editedticket', 'updated')}

        def produce((id, t, author, type, summary), status, fields, comment):
            if status == 'edit':
                if 'ticket_details' in filters:
                    info = ''
                    if len(fields) > 0:
                        info = ', '.join(['<i>%s</i>' % f for f in \
                                          fields.keys()]) + ' changed<br />'
                else:
                    return None
            elif 'ticket' in filters:
                if status == 'closed' and fields.has_key('resolution'):
                    info = fields['resolution']
                    if info and comment:
                        info = '%s: ' % info
                else:
                    info = ''
            else:
                return None
            kind, verb = status_map[status]
            title = util.Markup('Ticket <em title="%s">#%s</em> (%s) %s by %s',
                                summary, id, type, verb, author)
            href = format == 'rss' and self.env.abs_href.ticket(id) or \
                   self.env.href.ticket(id)

            if status == 'new':
                message = summary
            else:
                message = util.Markup(info)
                if comment:
                    if format == 'rss':
                        message += wiki_to_html(comment, self.env, req, db,
                                                absurls=True)
                    else:
                        message += wiki_to_oneliner(comment, self.env, db,
                                                    shorten=True)
            return kind, href, title, t, author, message

        # Ticket changes
        if 'ticket' in filters or 'ticket_details' in filters:
            db = self.env.get_db_cnx()
            cursor = db.cursor()

            priv_filter = ''
            if not req.authname or req.authname == 'anonymous':
                priv_filter = "AND t.status != 'unconfirmed' AND t.public = 1"

            cursor.execute("SELECT t.id,tc.time,tc.author,t.type,t.summary, "
                           "       tc.field,tc.oldvalue,tc.newvalue "
                           "  FROM ticket_change tc "
                           "    INNER JOIN ticket t ON t.id = tc.ticket "
                           "      AND tc.time>=%s AND tc.time<=%s %s "
                           "ORDER BY tc.time"
                           % (start, stop, priv_filter))
            previous_update = None
            for id,t,author,type,summary,field,oldvalue,newvalue in cursor:
                if not previous_update or (id,t,author) != previous_update[:3]:
                    if previous_update:
                        ev = produce(previous_update, status, fields, comment)
                        if ev:
                            yield ev
                    status, fields, comment = 'edit', {}, ''
                    previous_update = (id,t,author, type, summary)
                if field == 'comment':
                    comment = newvalue
                elif field == 'status' and newvalue in ('reopened', 'closed'):
                    status = newvalue
                else:
                    fields[field] = newvalue
            if previous_update:
                ev = produce(previous_update, status, fields, comment)
                if ev:
                    yield ev
            
            # New tickets
            if 'ticket' in filters:
                sql = "SELECT id,time,reporter,type,summary FROM ticket t WHERE time>=%s AND time<=%s " + priv_filter
                self.env.log.debug(sql % (start, stop))
                cursor.execute(sql, (start, stop))
                for row in cursor:
                    yield produce(row, 'new', {}, None)
 

    # Internal methods

    def _do_save(self, req, db, ticket, overlook_permissions = False):
        if not overlook_permissions == True:
            if req.perm.has_permission('TICKET_CHGPROP'):
                # TICKET_CHGPROP gives permission to edit the ticket
                if not req.args.get('summary'):
                    raise TracError('Tickets must contain summary.')
    
                if req.args.has_key('description') or req.args.has_key('reporter'):
                    req.perm.assert_permission('TICKET_ADMIN')
    
                ticket.populate(req.args)
            elif req.args.has_key('addcc'):
                req.perm.assert_permission('TICKET_ADD_CC')
            else:
                req.perm.assert_permission('TICKET_APPEND')

        # Mid air collision?
        if int(req.args.get('ts')) != ticket.time_changed:
            raise TracError("Sorry, can not save your changes. "
                            "This ticket has been modified by someone else "
                            "since you started", 'Mid Air Collision')

        # Do any action on the ticket?
        action = req.args.get('action')
        actions = TicketSystem(self.env).get_available_actions(ticket, req.perm)
        if action not in actions:
            raise TracError('Invalid action')
        
        author = req.args.get('author', req.authname) 

        # TODO: this should not be hard-coded like this
        if action == 'accept':
            ticket['status'] =  'assigned'
            ticket['owner'] = req.authname
        if action == 'resolve':
            ticket['status'] = 'closed'
            ticket['resolution'] = req.args.get('resolve_resolution')
        elif action == 'reassign':
            ticket['owner'] = req.args.get('reassign_owner')
            ticket['status'] = 'new'
        elif action == 'reopen':
            ticket['status'] = 'reopened'
            ticket['resolution'] = ''
        elif action == 'confirm':
            ticket['status'] = 'new'
            author = 'ticket system'
        elif action == 'addcc':
            email = util.validate_email(req.args.get('addcc'))
            if email is None:
                raise TracError("The email address you have entered doesn't "
                        "look like a valid email address. If this is in "
                        "error, please email staff@sccs to report this bug.",
                        'Invalid email')

            # try to send the confirmation email
            try:
                eml = TicketAddCcConfirmEmail(self.env)
                eml.notify(ticket, email)
            except Exception, e:
                self.log.exception("Failure sending notification on change to "
                        "ticket #%s: %s" % (ticket.id, e))


            raise TracMessage('A confirmation email has been sent to %s '
                    'containing instructions on how to finalize adding that '
                    'address to the "watchers" list for this ticket.' % 
                    email,
                    title = 'Email Address Submitted',
                    redirect_url=self.env.href.ticket(ticket.id))
        elif action == 'confirmcc':
            if 'cc' in ticket.values:
                ticket['cc'] += ', ' + req.args.get('addcc')
            else:
                ticket['cc'] = req.args.get('addcc')

            author = 'ticket system'


        now = int(time.time())
        ticket.save_changes(author, req.args.get('comment'), when=now, db=db)
        db.commit()

        try:
            if action == 'confirm':
                tn = TicketNotifyStaffEmail(self.env)
                tn.notify(ticket, newticket=True)
            elif not action == 'confirmcc':
                tn = TicketNotifyStaffEmail(self.env)
                tn.notify(ticket, newticket=False, modtime=now)
        except Exception, e:
            self.log.exception("Failure sending notification on change to "
                               "ticket #%s: %s" % (ticket.id, e))

        if action in ('resolve', 'reopen'):
            # only send the user notification emails when
            # the ticket is just closed or reopened
            try:
                tn = TicketNotifyEmail(self.env)
                tn.notify(ticket, newticket=False, modtime=now, allowedfields=['comment'])
            except Exception, e:
                self.log.exception("Failure sending notification on change to "
                                   "ticket #%s: %s" % (ticket.id, e))

        if action == 'confirmcc':
            raise TracMessage('This ticket has now been confirmed, and staff has been notified.',
                    'Email Address Confirmed',
                    redirect_url = self.env.href.ticket(ticket.id))
        elif action == 'confirm':
            raise TracMessage('This ticket has now been confirmed, and staff has been notified.',
                    'Ticket Confirmed',
                    redirect_url = self.env.href.ticket(ticket.id))
        else:
            req.redirect(self.env.href.ticket(ticket.id))

    def _insert_ticket_data(self, req, db, ticket, reporter_id):
        """Insert ticket data into the hdf"""
        req.hdf['ticket'] = ticket.values
        
        req.hdf['ticket.field.public'] = str( ticket['public'] == 1 )
        if not ticket['public']:
            req.hdf['ticket.field.type'] = ticket['type'] + ' [private]'

        req.hdf['ticket.id'] = ticket.id
        req.hdf['ticket.href'] = self.env.href.ticket(ticket.id)


        # transform the comma-separated list into a python list of strings
        userfields = self.env.config.get('trackle', 'user_fields')
        userfields = [field.strip() for field in userfields.replace(', ', ' ').split()]


        for field in TicketSystem(self.env).get_ticket_fields():
            if field['type'] in ('radio', 'select'):
                value = ticket.values.get(field['name'])
                options = field['options']
                if value and not value in options:
                    # Current ticket value must be visible even if its not in the
                    # possible values
                    options.append(value)
                field['options'] = options
            name = field['name']
            del field['name']
            if name in ('summary', 'reporter', 'description', 'type', 'status',
                        'resolution', 'owner'):
                field['skip'] = True

            if name not in userfields and not req.perm.has_permission('TICKET_VIEW_EXTENDED'):
                field['skip'] = True
                
            req.hdf['ticket.fields.' + name] = field

        req.hdf['ticket.reporter_id'] = reporter_id
        req.hdf['title'] = '#%d (%s)' % (ticket.id, ticket['summary'])
        req.hdf['ticket.description.formatted'] = wiki_to_html(ticket['description'],
                                                               self.env, req, db)

        req.hdf['ticket.opened'] = util.format_datetime(ticket.time_created)
        req.hdf['ticket.opened_delta'] = util.pretty_timedelta(ticket.time_created)
        if ticket.time_changed != ticket.time_created:
            req.hdf['ticket.lastmod'] = util.format_datetime(ticket.time_changed)
            req.hdf['ticket.lastmod_delta'] = util.pretty_timedelta(ticket.time_changed)
            req.hdf['ticket.lastmod_author'] = ticket.last_change_author

        changelog = ticket.get_changelog(db=db)
        curr_author = None
        curr_date   = 0
        changes = []
        for date, author, field, old, new in changelog:
            if date != curr_date or author != curr_author:
                changes.append({
                    'date': util.format_datetime(date),
                    'author': author,
                    'fields': {}
                })
                curr_date = date
                curr_author = author
            if field == 'comment':
                changes[-1]['comment'] = wiki_to_html(new, self.env, req, db)
            elif field == 'description':
                changes[-1]['fields'][field] = ''
            else:
                changes[-1]['fields'][field] = {'old': old,
                                                'new': new}
        req.hdf['ticket.changes'] = changes

        # List attached files
        for idx, attachment in enumerate(Attachment.select(self.env, 'ticket',
                                                           ticket.id)):
            hdf = attachment_to_hdf(self.env, db, req, attachment)
            req.hdf['ticket.attachments.%s' % idx] = hdf
        if req.perm.has_permission('TICKET_APPEND'):
            req.hdf['ticket.attach_href'] = self.env.href.attachment('ticket',
                                                                     ticket.id)

        # Add the possible actions to hdf
        actions = TicketSystem(self.env).get_available_actions(ticket, req.perm)
        for action in actions:
            req.hdf['ticket.actions.' + action] = '1'
