from trackle.db import Table, Column, Index, DatabaseManager
import time
import types
import pickle
import os, os.path

from trackle.wiki.model import WikiPage

import trackle.siteconfig

def do_upgrade(env, ver, cursor):
    """
    upates to shell session schema & data representation (stats stored separately, etc)
    """
    db_connector, _ = DatabaseManager(env)._get_connector()

    print 

    # create tables
    shellsession = Table('shellsession', key=('ticket', 'seq', 'version'))[
        Column('ticket', type='int'),
        Column('seq', type='int'),
        Column('version', type='int'),
        Column('author'),
        Column('ipnr'),
        Column('start', type='int'),
        Column('finish', type='int'),
        Column('text')]

    print "create table 'shellsession'...",
    for stmt in db_connector.to_sql(shellsession):
        cursor.execute(stmt)
    print "done"


    print "  > initial data migration...",
    cursor.execute("INSERT INTO shellsession (ticket,seq,version,author,ipnr,start,finish,text)"
                   "     SELECT CAST(ticket AS integer),time,version,author,ipnr,-1,-1,text"
                   "       FROM shelllogs")
    print "done (updated %d)" % (cursor.rowcount)
    print


    print "backup 'filediff' => 'filediff_old'...",
    cursor.execute("CREATE TABLE filediff_old AS SELECT * FROM filediff")
    print "done"
    print

    print "drop 'filediff'...",
    cursor.execute("DROP TABLE filediff")
    print "done"
    print

    filediff = Table('filediff', key=('ticket', 'seq', 'path'))[
        Column('ticket', type='int'),
        Column('seq', type='int'),
        Column('path'),
        Column('start'),
        Column('finish'),
        Column('sstat'),
        Column('fstat'),
        Column('active', type='int')]
        
    print "create (new) table 'filediff'...",
    for stmt in db_connector.to_sql(filediff):
        cursor.execute(stmt)
    print "done"


    # we use the made-up 'nostat' value in the special field, since we have no
    # stat information about these files to migrate to the new tables
    print "  > initial data migration...",
    cursor.execute("INSERT INTO filediff (ticket, seq, path, start, finish, active) "
                   "     SELECT CAST( substr(id, 0, strpos(id, '-'))      AS integer), "
                   "            CAST( substr(id,    strpos(id, '-') + 1)  AS integer), "
                   "            filename, fdiff_start, fdiff_finish, active "
                   "       FROM filediff_old")
    print "done (updated %d)" % (cursor.rowcount)
    print


    

    # now migrate the old times to sequence nums
    print "migrating times in 'shellsession', 'filediff' to sequence numbers..."


    print "  > creating unique sequence number for each shell session"

    # this result set returns one row per logical shell
    # session, regardless of number of revisions of
    # session wiki text area
    cursor.execute("SELECT DISTINCT ticket,time FROM shelllogs ORDER BY ticket ASC, time ASC")
    result = cursor.fetchall()


    # this will map old shell session IDs ("<ticket>-<time>")
    # to the sequence number of the shell session relative
    # to its ticket ("<seq>")
    tidToSeq = {}

    lastTicket = -1
    seqCounter = 1

    for ticket,tim in result:

        if ticket != lastTicket:
            # this is the first time we've seen this ticket
            # in this result set, so reset seq counter to 1
            seqCounter = 1
            lastTicket = ticket

        tid = '%s-%s' % (ticket, tim)
        tidToSeq[tid] = seqCounter
        seqCounter += 1

    

    print "  > updating shellsession and filediff with session sequences"
    # also parsing start and end dates for the shell sessions
    cursor.execute("SELECT ticket,time,version,start,finish FROM shelllogs ORDER BY ticket ASC, time ASC")
    result = cursor.fetchall()


    timefmt = '%a %b %d %H:%M:%S %Z %Y'

    # eventually saved and used in a separate step
    # to automatically update links in wiki areas, etc
    map = {}

    for ticket,tim,version,start,finish in result:

        tid = '%s-%s' % (ticket, tim)
        seq = tidToSeq[tid]

        ticket = int(ticket)

        # update the map
        newtid = '%s-%s' % (ticket, seq)
        map[tid] = newtid

        start  = int(time.mktime(time.strptime(start.strip(),  timefmt)))
        finish = int(time.mktime(time.strptime(finish.strip(), timefmt)))

        args  = (seq, start, finish, ticket, tim, version)
        args2 = (seq,                ticket, tim)

        print "  > shellsession '%s' => '%s'..." % (tid, newtid),
        cursor.execute("UPDATE shellsession SET seq=%d, start=%d, finish=%d WHERE ticket=%d AND seq=%d AND version=%d", args)
        print "done (updated %d)" % (cursor.rowcount)

        print "  > filediff(s)  '%s' => '%s'..." % (tid, newtid),
        cursor.execute("UPDATE filediff     SET seq=%d                      WHERE ticket=%d AND seq=%d", args2)
        print "done (updated %d)" % (cursor.rowcount)
    print


    # now we have to extract the diff info and update the new table
    print "extracting stat info from diff'd stats..."
    cursor.execute("SELECT id, filename, sdiff FROM filediff_old")
    result = cursor.fetchall()

    for id, path, statdiff in result:
        pre, post = diffsplit(statdiff)

        (ticket, seq) = map[id].split('-')

        args = (pre, post, int(ticket), int(seq), path)
        print "  > updating stat for '%s' in session '%s-%s'..." % (path, ticket, seq),
        cursor.execute("UPDATE filediff SET sstat=%s, fstat=%s WHERE ticket=%d AND seq=%d AND path=%s", args)
        print "done (updated %d)" % (cursor.rowcount)
    print

    print "saving shell session sequence migration map...",
    fp = file( os.path.join( trackle.siteconfig.__default_conf_dir__, 'shellsession-migr.map'), 'w')
    pickle.dump( map, fp )
    fp.close()
    print "done"
    print






    print """This database upgrade changes the way Trackle stores and links to
shell session logs. As such, you may now have dead links in editable
areas of the Trackle website (ticket descriptions and discussion, wiki
pages, report headers, milestones, other shell session logs, etc. This
upgrade script has only modified the shell sessions portion of the
database -- to fix broken links in the editable areas, you can use
the "shellsession fixlinks" action of trackle-admin:

    trackle-admin shellsession fixlinks

This will create new versions for versioned resources (wiki pages,
shell session logs, etc).


Additionally, the old tables 'shelllogs' and 'filediff' have been saved
in the database as 'shelllogs' and 'filediff_old', respectively. Once
you have verified the migration was succesful, you may remove these with
trackle-admin:

    trackle-admin db cleanup shelllogs
"""




def diffsplit(diff):
    """
    take a diff and extract the pre- and post- files
    from it, as much as possible (if lines are missing
    then we can't reconstruct them)

    return (pre, post) as newline-joined strings. the
    strings will have the 'special' field
    """

    if 'REMOVED' in diff:
        return ('special: ', 'special: noexist')

    elif 'CREATED' in diff:
        return ('special: noexist', 'special: ')
    
    pre = []
    post = []

    if type(diff) in types.StringTypes:
        diff = diff.splitlines()

    for line in diff:
        if line[:2] in ('++', '--', '@@'):
            continue

        if line[0] == ' ':
            pre.append(line.strip())
            post.append(line.strip())

        elif line[0] == '+':
            post.append(line[1:].strip())

        elif line[0] == '-':
            pre.append(line[1:].strip())

    pre.append('special: ')
    post.append('special: ')

    pre = '\n'.join(pre)
    post = '\n'.join(post)

    return (pre, post)
